"""
project
=======

At the core of scPortrait is the concept of a `Project`. A `Project` is a Python class that orchestrates all scPortrait processing steps, serving as the central element for all operations.
Each `Project` corresponds to a directory on the file system, which houses the input data for a specific scPortrait run along with the generated outputs.
The choice of the appropriate `Project` class depends on the structure of the data to be processed.

For more details, refer to :ref:`here <projects>`.
"""

from __future__ import annotations

import os
import re
import shutil
import tempfile
from pathlib import PosixPath
from typing import TYPE_CHECKING, Literal

import dask.array as da
import dask.array as darray
import matplotlib.pyplot as plt
import numpy as np
import psutil
import zarr
from alphabase.io import tempmmap
from spatialdata import SpatialData
from tifffile import imread

from scportrait.io import daskmmap
from scportrait.pipeline._base import Logable
from scportrait.pipeline._utils.helper import _check_for_spatialdata_plot, read_config, write_config
from scportrait.pipeline._utils.sdata_io import sdata_filehandler
from scportrait.pipeline._utils.spatialdata_helper import (
    get_chunk_size,
    rechunk_image,
)
from scportrait.tools.sdata.write._helper import _get_image, _get_shape

if TYPE_CHECKING:
    from collections.abc import Callable

    from anndata import AnnData
    from matplotlib.pyplot import Figure

from scportrait.io import read_h5sc
from scportrait.pipeline._utils.constants import (
    DEFAULT_CELL_ID_NAME,
    DEFAULT_CENTERS_NAME,
    DEFAULT_CHUNK_SIZE_2D,
    DEFAULT_CHUNK_SIZE_3D,
    DEFAULT_CONFIG_NAME,
    DEFAULT_DATA_DIR,
    DEFAULT_EXTRACTION_DIR_NAME,
    DEFAULT_EXTRACTION_FILE,
    DEFAULT_FEATURIZATION_DIR_NAME,
    DEFAULT_IMAGE_DTYPE,
    DEFAULT_INPUT_IMAGE_NAME,
    DEFAULT_PREFIX_FILTERED_SEG,
    DEFAULT_PREFIX_MAIN_SEG,
    DEFAULT_PREFIX_SELECTED_SEG,
    DEFAULT_SDATA_FILE,
    DEFAULT_SEG_NAME_0,
    DEFAULT_SEG_NAME_1,
    DEFAULT_SEGMENTATION_DIR_NAME,
    DEFAULT_SEGMENTATION_DTYPE,
    DEFAULT_SELECTION_DIR_NAME,
    DEFAULT_SINGLE_CELL_IMAGE_DTYPE,
    ChunkSize2D,
    ChunkSize3D,
)
from scportrait.processing.images._image_processing import percentile_normalization


class Project(Logable):
    """Base implementation for a scPortrait ``project``.

    This class is designed to handle single-timepoint, single-location data, like e.g. whole-slide images.

    Segmentation Methods should be based on :func:`Segmentation <scportrait.pipeline.segmentation.Segmentation>` or :func:`ShardedSegmentation <scportrait.pipeline.segmentation.ShardedSegmentation>`.
    Extraction Methods should be based on :func:`HDF5CellExtraction <scportrait.pipeline.extraction.HDF5CellExtraction>`.

    Attributes:
        config (dict): Dictionary containing the config file.
        nuc_seg_name (str): Name of the nucleus segmentation object.
        cyto_seg_name (str): Name of the cytosol segmentation object.
        sdata_path (str): Path to the spatialdata object.
        filehander (sdata_filehandler): Filehandler for the spatialdata object which manages all calls or updates to the spatialdata object.
    """

    # define cleanup behaviour
    CLEAN_LOG: bool = True

    # import all default values from constants
    DEFAULT_CONFIG_NAME = DEFAULT_CONFIG_NAME
    DEFAULT_INPUT_IMAGE_NAME = DEFAULT_INPUT_IMAGE_NAME
    DEFAULT_SDATA_FILE = DEFAULT_SDATA_FILE

    DEFAULT_PREFIX_MAIN_SEG = DEFAULT_PREFIX_MAIN_SEG
    DEFAULT_PREFIX_FILTERED_SEG = DEFAULT_PREFIX_FILTERED_SEG
    DEFAULT_PREFIX_SELECTED_SEG = DEFAULT_PREFIX_SELECTED_SEG

    DEFAULT_SEG_NAME_0: str = DEFAULT_SEG_NAME_0
    DEFAULT_SEG_NAME_1: str = DEFAULT_SEG_NAME_1

    DEFAULT_CENTERS_NAME: str = DEFAULT_CENTERS_NAME

    DEFAULT_CHUNK_SIZE_3D: ChunkSize3D = DEFAULT_CHUNK_SIZE_3D
    DEFAULT_CHUNK_SIZE_2D: ChunkSize2D = DEFAULT_CHUNK_SIZE_2D

    DEFAULT_SEGMENTATION_DIR_NAME = DEFAULT_SEGMENTATION_DIR_NAME
    DEFAULT_EXTRACTION_DIR_NAME = DEFAULT_EXTRACTION_DIR_NAME
    DEFAULT_DATA_DIR = DEFAULT_DATA_DIR
    DEFAULT_EXTRACTION_FILE = DEFAULT_EXTRACTION_FILE

    DEFAULT_FEATURIZATION_DIR_NAME = DEFAULT_FEATURIZATION_DIR_NAME
    DEFAULT_SELECTION_DIR_NAME = DEFAULT_SELECTION_DIR_NAME

    DEFAULT_IMAGE_DTYPE = DEFAULT_IMAGE_DTYPE
    DEFAULT_SEGMENTATION_DTYPE = DEFAULT_SEGMENTATION_DTYPE
    DEFAULT_SINGLE_CELL_IMAGE_DTYPE = DEFAULT_SINGLE_CELL_IMAGE_DTYPE
    DEFAULT_CELL_ID_NAME = DEFAULT_CELL_ID_NAME

    PALETTE = [
        "blue",
        "green",
        "red",
        "yellow",
        "purple",
        "orange",
        "pink",
        "cyan",
        "magenta",
        "lime",
        "teal",
        "lavender",
        "brown",
        "beige",
        "maroon",
        "mint",
        "olive",
        "apricot",
        "navy",
        "grey",
        "white",
        "black",
    ]

    def __init__(
        self,
        project_location: str,
        config_path: str = None,
        segmentation_f=None,
        extraction_f=None,
        featurization_f=None,
        selection_f=None,
        overwrite: bool = False,
        debug: bool = False,
    ):
        """
        Args:
            project_location (str): Path to the project directory.
            config_path (str): Path to the config file.
            segmentation_f (optional): Segmentation method to be used for the project.
            extraction_f (optional): Extraction method to be used for the project.
            featurization_f (optional): Featurization method to be used for the project.
            selection_f (optional): Selection method to be used for the project.
            overwrite (optional): If set to ``True``, will overwrite existing files in the project directory. Default is ``False``.
            debug (optional): If set to ``True``, will print additional debug messages/plots. Default is ``False``.
        """
        super().__init__(directory=project_location, debug=debug)

        self.project_location = project_location
        self.overwrite = overwrite
        self.config: None | dict = None
        if config_path is None or isinstance(config_path, str | PosixPath):
            self._get_config_file(config_path)
        elif isinstance(config_path, dict):
            self._load_config_from_dict(config_path)

        self.nuc_seg_name = f"{self.DEFAULT_PREFIX_MAIN_SEG}_{self.DEFAULT_SEG_NAME_0}"
        self.cyto_seg_name = f"{self.DEFAULT_PREFIX_MAIN_SEG}_{self.DEFAULT_SEG_NAME_1}"

        self.segmentation_f = segmentation_f
        self.extraction_f = extraction_f
        self.featurization_f = featurization_f
        self.selection_f = selection_f

        # intialize containers to support interactive viewing of the spatialdata object
        self.interactive_sdata = None
        self.interactive = None

        if self.CLEAN_LOG:
            self._clean_log_file()

        # check if project directory exists, if it does not create
        if not os.path.isdir(self.project_location):
            os.makedirs(self.project_location)
        else:
            Warning("There is already a directory in the location path")

        # === setup sdata reader/writer ===
        self.filehandler = sdata_filehandler(
            directory=self.directory,
            sdata_path=self.sdata_path,
            input_image_name=self.DEFAULT_INPUT_IMAGE_NAME,
            nuc_seg_name=self.nuc_seg_name,
            cyto_seg_name=self.cyto_seg_name,
            centers_name=self.DEFAULT_CENTERS_NAME,
            default_cell_id_name=self.DEFAULT_CELL_ID_NAME,
            debug=self.debug,
        )
        self.filehandler._check_sdata_status()  # update sdata object and status

        # === setup segmentation ===
        self._setup_segmentation_f(segmentation_f)

        # === setup extraction ===
        self._setup_extraction_f(extraction_f)

        # === setup featurization ===
        self._setup_featurization_f(featurization_f)

        # ==== setup selection ===
        self._setup_selection(selection_f)

    def __exit__(self):
        self._clear_temp_dir()

    def __del__(self):
        self._clear_temp_dir()

    @property
    def sdata_path(self) -> str:
        return self._get_sdata_path()

    @property
    def sdata(self) -> SpatialData:
        """Shape of data matrix (:attr:`n_obs`, :attr:`n_vars`)."""
        return self.filehandler.get_sdata()

    @property
    def h5sc(self) -> AnnData:
        if self.extraction_f is None:
            raise ValueError("No extraction method has been set.")
        else:
            if self.extraction_f.output_path is None:
                path = self.extraction_f.extraction_file
            else:
                path = self.extraction_f.output_path
            return read_h5sc(path)

    ##### Setup Functions #####

    def _load_config_from_file(self, file_path):
        """
        Loads config from file and writes it to self.config

        Args:
            file_path (str): Path to the config.yml file that should be loaded.
        """
        self.log(f"Loading config from {file_path}")

        if not os.path.isfile(file_path):
            raise ValueError(f"Your config path {file_path} is invalid.")

        self.config = read_config(file_path)

    def _load_config_from_dict(self, config_dict):
        self.config = config_dict
        write_config(self.config, os.path.join(self.project_location, self.DEFAULT_CONFIG_NAME))

    def _get_config_file(self, config_path: str | None = None) -> None:
        """Load the config file for the project. If no config file is passed the default config file in the project directory is loaded.

        Args:
            config_path (str, optional): Path to the config file. Default is ``None``.

        Returns:
            None: the config dictionary project.config is updated.
        """
        # load config file
        self.config_path = os.path.join(self.project_location, self.DEFAULT_CONFIG_NAME)

        if config_path is None:
            # Check if there is already a config file in the dataset folder in case no config file has been specified

            if os.path.isfile(self.config_path):
                self._load_config_from_file(self.config_path)
            else:
                raise ValueError("No config passed and no config found in project directory.")

        else:
            if not os.path.isfile(config_path):
                raise ValueError(f"Your config path {config_path} is invalid. Please specify a valid config path.")

            else:
                print("Updating project config file.")

                if os.path.isfile(self.config_path):
                    os.remove(self.config_path)

                # ensure that the project location exists
                if not os.path.isdir(self.project_location):
                    os.makedirs(self.project_location)

                # The blueprint config file is copied to the dataset folder and renamed to the default name
                shutil.copy(config_path, self.config_path)
                self._load_config_from_file(self.config_path)

    def _setup_segmentation_f(self, segmentation_f):
        """Configure the segmentation method for the project.

        Args:
            segmentation_f (Callable): Segmentation method to be used for the project.

        Returns:
            None: the segmentation method is updated in the project object.
        """
        if segmentation_f is not None:
            if segmentation_f.__name__ not in self.config:
                raise ValueError(f"Config for {segmentation_f.__name__} is missing from the config file.")

            seg_directory = os.path.join(self.project_location, self.DEFAULT_SEGMENTATION_DIR_NAME)

            self.seg_directory = seg_directory

            self.segmentation_f = segmentation_f(
                self.config[segmentation_f.__name__],
                self.seg_directory,
                nuc_seg_name=self.nuc_seg_name,
                cyto_seg_name=self.cyto_seg_name,
                _tmp_image_path=None,
                project_location=self.project_location,
                debug=self.debug,
                overwrite=self.overwrite,
                project=None,
                filehandler=self.filehandler,
                from_project=True,
            )

    def _update_segmentation_f(self, segmentation_f):
        self._setup_segmentation_f(segmentation_f)

    def _setup_extraction_f(self, extraction_f):
        """Configure the extraction method for the project.

        Args:
            extraction_f (Callable): Extraction method to be used for the project.

        Returns:
            None: the extraction method is updated in the project object.
        """

        if extraction_f is not None:
            extraction_directory = os.path.join(self.project_location, self.DEFAULT_EXTRACTION_DIR_NAME)

            self.extraction_directory = extraction_directory

            if extraction_f.__name__ not in self.config:
                raise ValueError(f"Config for {extraction_f.__name__} is missing from the config file")

            self.extraction_f = extraction_f(
                self.config[extraction_f.__name__],
                self.extraction_directory,
                project_location=self.project_location,
                debug=self.debug,
                overwrite=self.overwrite,
                project=self,
                filehandler=self.filehandler,
                from_project=True,
            )

    def _update_extraction_f(self, extraction_f):
        self._setup_extraction_f(extraction_f)

    def _setup_featurization_f(self, featurization_f):
        """Configure the featurization method for the project.

        Args:
            featurization_f (Callable): Featurization method to be used for the project.

        Returns:
            None: the featurization method is updated in the project object.
        """
        if featurization_f is not None:
            if featurization_f.__name__ not in self.config:
                raise ValueError(f"Config for {featurization_f.__name__} is missing from the config file")

            featurization_directory = os.path.join(self.project_location, self.DEFAULT_FEATURIZATION_DIR_NAME)

            self.featurization_directory = featurization_directory

            self.featurization_f = featurization_f(
                self.config[featurization_f.__name__],
                self.featurization_directory,
                project_location=self.project_location,
                debug=self.debug,
                overwrite=False,  # this needs to be set to false as the featurization step should not remove previously created features
                project=self,
                filehandler=self.filehandler,
                from_project=True,
            )

    def update_featurization_f(self, featurization_f):
        """Update the featurization method chosen for the project without reinitializing the entire project.

        Args:
            featurization_f : The featurization method that should be used for the project.

        Returns:
            None : the featurization method is updated in the project object.

        Examples:
            Update the featurization method for a project::

                from scportrait.pipeline.featurization import CellFeaturizer

                project.update_featurization_f(CellFeaturizer)
        """
        self.log(f"Replacing current featurization method {self.featurization_f.__class__} with {featurization_f}")
        self._setup_featurization_f(featurization_f)

    def _setup_selection(self, selection_f):
        """Configure the selection method for the project.

        Args:
            selection_f (Callable): Selection method to be used for the project.

        Returns:
            None: the selection method is updated in the project object.
        """
        if self.selection_f is not None:
            if selection_f.__name__ not in self.config:
                raise ValueError(f"Config for {selection_f.__name__} is missing from the config file")

            selection_directory = os.path.join(self.project_location, self.DEFAULT_SELECTION_DIR_NAME)

            self.selection_directory = selection_directory

            self.selection_f = selection_f(
                self.config[selection_f.__name__],
                self.selection_directory,
                project_location=self.project_location,
                debug=self.debug,
                overwrite=self.overwrite,
                project=self,
                filehandler=self.filehandler,
                from_project=True,
            )

    def _update_selection_f(self, selection_f):
        self._setup_selection(selection_f)

    ##### General small helper functions ####

    def _check_memory(self, item):
        """
        Check the memory usage of the given if it were completely loaded into memory using .compute().
        """
        array_size = item.nbytes
        available_memory = psutil.virtual_memory().available

        return array_size < available_memory

    def _check_chunk_size(self, elem, chunk_size):
        """
        Check if the chunk size of the element is the default chunk size. If not rechunk the element to the default chunk size.
        """

        # get chunk size of element
        chunk_size = get_chunk_size(elem)

        if isinstance(chunk_size, list):
            # check if all chunk sizes are the same otherwise rechunking needs to occur anyways
            if not all(x == chunk_size[0] for x in chunk_size):
                elem = rechunk_image(elem, chunk_size=chunk_size)
            else:
                # ensure that the chunk size is the default chunk size
                if chunk_size != chunk_size:
                    elem = rechunk_image(elem, chunk_size=chunk_size)
        else:
            # ensure that the chunk size is the default chunk size
            if chunk_size != chunk_size:
                elem = rechunk_image(elem, chunk_size=chunk_size)

    def _check_image_dtype(self, image: np.ndarray) -> None:
        """Check if the image dtype is the default image dtype.

        Args:
            image (np.ndarray): Image to be checked.

        Returns:
            None: If the image dtype is the default image dtype, no action is taken.

        Raises:
            Warning: If the image dtype is not the default image dtype.
        """

        if not image.dtype == self.DEFAULT_IMAGE_DTYPE:
            Warning(
                f"Image dtype is not {self.DEFAULT_IMAGE_DTYPE} but insteadt {image.dtype}. The workflow expects images to be of dtype {self.DEFAULT_IMAGE_DTYPE}. Proceeding with the incorrect dtype can lead to unexpected results."
            )
            self.log(
                f"Image dtype is not {self.DEFAULT_IMAGE_DTYPE} but insteadt {image.dtype}. The workflow expects images to be of dtype {self.DEFAULT_IMAGE_DTYPE}. Proceeding with the incorrect dtype can lead to unexpected results."
            )

    def _create_temp_dir(self, path) -> None:
        """
        Create a temporary directory in the specified directory with the name of the class.

        Args:
            path (str): Path to the directory where the temporary directory should be created.

        Returns:
            None: The temporary directory is created in the specified directory. The path to the temporary directory is stored in the project object as self._tmp_dir_path.

        """

        path = os.path.join(path, f"{self.__class__.__name__}_")
        self._tmp_dir = tempfile.TemporaryDirectory(prefix=path)
        self._tmp_dir_path = self._tmp_dir.name
        """str: Path to the temporary directory."""

        self.log(f"Initialized temporary directory at {self._tmp_dir_path} for {self.__class__.__name__}")

    def _clear_temp_dir(self):
        """Clear the temporary directory."""
        if "_tmp_dir" in self.__dict__.keys():
            shutil.rmtree(self._tmp_dir_path, ignore_errors=True)
            self.log(f"Cleaned up temporary directory at {self._tmp_dir}")

            del self._tmp_dir, self._tmp_dir_path
        else:
            self.log("Temporary directory not found, skipping cleanup")

    ##### Functions for handling sdata object #####

    def _cleanup_sdata_object(self):
        """
        Check if the output location exists and if it does cleanup if allowed, otherwise raise an error.
        """

        if os.path.exists(self.sdata_path):
            if self.overwrite:
                if not self.filehandler._check_empty_sdata():
                    self.log(f"Output location {self.sdata_path} already exists. Overwriting.")
                    shutil.rmtree(self.sdata_path, ignore_errors=True)
            else:
                # check to see if the sdata object is empty
                if not self.filehandler._check_empty_sdata():
                    raise ValueError(
                        f"Output location {self.sdata_path} already exists. Set overwrite=True to overwrite."
                    )
        self.filehandler._check_sdata_status()

    def _get_sdata_path(self):
        """
        Get the path to the spatialdata object.
        """
        return os.path.join(self.project_location, self.DEFAULT_SDATA_FILE)

    def print_project_status(self):
        """Print the current project status."""
        self.get_project_status(print_status=True)

    def get_project_status(self, print_status=False):
        self.filehandler._check_sdata_status()
        self.input_image_status = self.filehandler.input_image_status
        self.nuc_seg_status = self.filehandler.nuc_seg_status
        self.cyto_seg_status = self.filehandler.cyto_seg_status
        self.nuc_centers_status = self.filehandler.nuc_centers_status
        self.cyto_centers_status = self.filehandler.cyto_centers_status
        extraction_file = os.path.join(
            self.project_location, self.DEFAULT_EXTRACTION_DIR_NAME, self.DEFAULT_DATA_DIR, self.DEFAULT_EXTRACTION_FILE
        )
        self.extraction_status = True if os.path.isfile(extraction_file) else False

        if self.input_image_status:
            if self.DEFAULT_INPUT_IMAGE_NAME in self.sdata:
                self.input_image = _get_image(self.sdata[self.DEFAULT_INPUT_IMAGE_NAME])
            else:
                self.input_image = None

        if print_status:
            self.log("Current Project Status:")
            self.log("--------------------------------")
            self.log(f"Input Image in sdata: {self.input_image_status}")
            self.log(f"Nucleus Segmentation in sdata: {self.nuc_seg_status}")
            self.log(f"Cytosol Segmentation in sdata: {self.cyto_seg_status}")
            self.log(f"Nucleus Centers in sdata: {self.nuc_centers_status}")
            self.log(f"Cytosol Centers in sdata: {self.cyto_centers_status}")
            self.log(f"Extracted single-cell images saved to file: {self.extraction_status}")
            self.log("--------------------------------")

        return None

    def view_sdata(self):
        """Start an interactive napari viewer to look at the sdata object associated with the project.
        Note:
            This only works in sessions with a visual interface.
        """
        # open interactive viewer in napari
        try:
            from napari_spatialdata import Interactive
        except ImportError:
            raise ImportError(
                "napari-spatialdata must be installed to use the interactive viewer. Please install with `pip install scportrait[plotting]`."
            ) from None
        self.interactive_sdata = self.filehandler.get_sdata()
        self.interactive = Interactive(self.interactive_sdata)
        self.interactive.run()

    def _save_interactive_sdata(self):
        assert self.interactive_sdata is not None, "No interactive sdata object found."

        in_memory_only, _ = self.interactive_sdata._symmetric_difference_with_zarr_store()
        print(f"Writing the following manually added files to the sdata object: {in_memory_only}")

        dict_lookup = {}
        for elem in in_memory_only:
            key, name = elem.split("/")
            if key not in dict_lookup:
                dict_lookup[key] = []
            dict_lookup[key].append(name)
        for _, name in dict_lookup.items():
            self.interactive_sdata.write_element(name)  # replace with correct function once pulled in from sdata

    def close_interactive_viewer(self):
        if self.interactive is not None:
            self._save_interactive_sdata()
            self.interactive._viewer.close()

            # reset to none values to track next call of view_sdata
            self.interactive_sdata = None
            self.interactive = None

    def _check_for_interactive_session(self):
        if self.interactive is not None:
            Warning("Interactive viewer is still open. Will automatically close before proceeding with processing.")
            self.close_interactive_viewer()

    #### Functions to visualize results ####

    def plot_input_image(
        self,
        image_name="input_image",
        max_width: int = 1000,
        select_region: tuple[int, int] | None = None,
        channels: list[int] | list[str] | None = None,
        normalize: bool = False,
        normalization_percentile: tuple[float, float] = (0.01, 0.99),
        title_fontsize: int = 20,
        figsize_single_tile: tuple[float, float] = (8, 8),
        return_fig: bool = False,
    ) -> Figure | None:
        """Plot the input image associated with the project. If the image is large it will automatically plot a subset in the center

        Args:
            image_name: Name of the element containing the input image in the spatialdata object.
            max_width: Maximum size of the image to be plotted in pixels.
            select_region: Tuple containing the x and y coordinates of the center of the region to be plotted. If not set it will use the center of the image.
            channels: List of channel names or indices to be plotted. If not set, the first 4 channels will be plotted.
            normalize: If set to ``True``, the image will be normalized to the specified percentile values.
            normalization_percentile: Tuple containing the lower and upper percentile values to be used for normalization.
            title_fontsize: Fontsize of the title of the plot.
            figsize_single_tile: size to be used during figure creation for a single axes in the resulting plot.
            return_fig: If set to ``True``, the function returns the figure object instead of displaying it.

        Returns:
            A matplotlib figure object if return_fig is set to ``True``.

        Examples:
            Plot the input image of a project::

                project.plot_input_image()
        """

        # check if spatialdata_plot is installed
        _check_for_spatialdata_plot()

        from scportrait.plotting import plot_image
        from scportrait.tools.sdata.processing import get_bounding_box_sdata

        _sdata = self.sdata

        # get relevant information from spatialdata object
        c, x, y = _sdata[image_name].scale0.image.shape
        channel_names = list(_sdata[image_name].scale0.c.values)

        if max_width is not None:
            # get center coordinates
            if select_region is None:
                center_x = x // 2
                center_y = y // 2
            else:
                center_x, center_y = select_region

            # subset spatialdata object if its too large
            if x > max_width or y > max_width:
                _sdata = get_bounding_box_sdata(_sdata, max_width, center_x, center_y)

        if normalize:
            lower_percentile, upper_percentile = normalization_percentile

            # get percentile values to normalize viewing to
            for channel in channel_names:
                idx = list(_sdata[image_name].scale0.c.values).index(channel)
                for scale in _sdata[image_name]:
                    im = _sdata[image_name].get(scale).image[idx].compute()
                    _sdata[image_name].get(scale).image[idx] = (
                        percentile_normalization(im, lower_percentile, upper_percentile) * np.iinfo(np.uint16).max
                    ).astype(np.uint16)

        if channels is not None:
            if isinstance(channels[0], int):
                assert all(
                    x in range(c) for x in channels
                ), "The specified channel indices are not found in the spatialdata object."
                valid_channels = [i for i in channels if isinstance(i, int)]
                channel_names = [channel_names[i] for i in valid_channels]
            if isinstance(channels[0], str):
                assert all(
                    x in channel_names for x in channels
                ), "The specified channel names are not found in the spatialdata object."
                channel_names = channels

            c = len(channels)
            palette = self.PALETTE[:c]
        else:
            palette = self.PALETTE[:c]

        fig_size_x, fig_size_y = figsize_single_tile
        fig, axs = plt.subplots(1, len(channel_names) + 1, figsize=(fig_size_x * (len(channel_names) + 1), fig_size_y))

        plot_image(
            _sdata,
            image_name=image_name,
            channel_names=channel_names,
            palette=palette,
            ax=axs[0],
            title="overlayed",
            title_fontsize=title_fontsize,
            return_fig=False,
            show_fig=False,
        )

        for i, channel in enumerate(channel_names):
            plot_image(
                _sdata,
                image_name=image_name,
                channel_names=[channel],
                palette=[palette[i]],
                ax=axs[i + 1],
                title=channel,
                title_fontsize=title_fontsize,
                return_fig=False,
                show_fig=False,
            )

        fig.tight_layout()
        if return_fig:
            return fig
        else:
            plt.show()
            return None

    def plot_he_image(
        self,
        image_name: str = "he_image",
        max_width: int | None = None,
        select_region: tuple[int, int] | None = None,
        title_fontsize: int = 20,
        return_fig: bool = False,
    ) -> None | Figure:
        """Plot the hematoxylin and eosin (HE) channel of the input image.

        Args:
            image_name: Name of the element containing the H&E image in the spatialdata object.
            max_width: Maximum width of the image to be plotted in pixels.
            select_region: Tuple containing the x and y coordinates of the region to be plotted. If not set it will use the center of the image.
            title_fontsize: Fontsize of the title of the plot.
            return_fig: If set to ``True``, the function returns the figure object instead of displaying it.

        Returns:
            A matplotlib figure object if return_fig is set to ``True``.

        Examples:
            Plot the HE channel of a project::

                project.plot_he()
        """

        _check_for_spatialdata_plot()

        # import optional dependencies required for this method
        import spatialdata_plot  # this does not have an explicit call put allows for sdata.pl calls

        _sdata = self.sdata

        # remove points object as this makes it
        points_keys = list(_sdata.points.keys())
        if len(points_keys) > 0:
            for x in points_keys:
                del _sdata.points[x]

        channel_names = list(_sdata[image_name].scale0.c.values)
        assert channel_names == ["r", "g", "b"], "The image is not an RGB image."

        # subset spatialdata object if its too large
        if max_width is not None:
            c, x, y = _sdata[image_name].scale0.image.shape
            width = max_width // 2
            if select_region is None:
                center_x = x // 2
                center_y = y // 2
            else:
                center_x, center_y = select_region

            if x > max_width or y > max_width:
                _sdata = _sdata.query.bounding_box(
                    axes=["x", "y"],
                    min_coordinate=[center_x - width, center_y - width],
                    max_coordinate=[center_x + width, center_y + width],
                    target_coordinate_system="global",
                )

        fig, ax = plt.subplots(1, 1, figsize=(8, 8))
        _sdata.pl.render_images(image_name).pl.show(ax=ax)
        ax.axis("off")
        ax.set_title("H&E Image", fontsize=title_fontsize)

        fig.tight_layout()

        if return_fig:
            return fig
        plt.show()
        return None

    def plot_segmentation_masks(
        self,
        max_width: int = 1500,
        select_region: tuple[int, int] | None = None,
        image_name: str = "input_image",
        mask_names: list[str] | None = None,
        normalize: bool = False,
        normalization_percentile: tuple[float, float] = (0.01, 0.99),
        title_fontsize: int = 20,
        line_width: int = 1,
        return_fig: bool = False,
    ) -> None | Figure:
        """Plot the generated segmentation masks. If the image is large it will automatically plot a subset cropped to the center of the spatialdata object.

        Uses the function `scportrait.plotting.sdata.plot_segmentation_mask` to plot the found segmentation masks in an scPortrait project.
        Please refer to the documentation of this function for further customization.

        Args:
            return_fig: If set to ``True``, the function returns the figure object instead of displaying it.
            max_width: Maximum width of the image to be plotted in pixels.
            select_region: Tuple containing the x and y coordinates of the region to be plotted. If not set it will use the center of the image.
            image_name: Name of the element containing the input image in the spatialdata object.
            mask_names: List of mask names to be plotted. If not set, all available masks will be plotted.
            normalize: If set to ``True``, the image will be normalized to the specified percentile values.
            normalization_percentile: Tuple containing the lower and upper percentile values to be used for normalization.
            title_fontsize: Fontsize of the title of the plot.
            line_width: Width of the lines in the plot.
            return_fig: If set to ``True``, the function returns the figure object instead of displaying it.

        Returns:
            A matplotlib figure object if return_fig is set to ``True``.

        Examples:
            Plot the segmentation masks of a project::

                project.plot_segmentation_masks()
        """
        _check_for_spatialdata_plot()

        # import relevant functions for this method
        from scportrait.plotting.sdata import plot_segmentation_mask
        from scportrait.tools.sdata.processing import get_bounding_box_sdata

        _sdata = self.sdata

        # get relevant information from spatialdata object
        _, x, y = _sdata[image_name].scale0.image.shape
        channel_names = list(_sdata[image_name].scale0.c.values)

        if max_width is not None:
            # get center coordinates
            if select_region is None:
                center_x = x // 2
                center_y = y // 2
            else:
                center_x, center_y = select_region

            # subset spatialdata object if its too large
            if x > max_width or y > max_width:
                _sdata = get_bounding_box_sdata(_sdata, max_width, center_x, center_y)

        if normalize:
            lower_percentile, upper_percentile = normalization_percentile

            # get percentile values to normalize viewing to
            for channel in channel_names:
                idx = list(_sdata[image_name].scale0.c.values).index(channel)
                for scale in _sdata[image_name]:
                    im = _sdata[image_name].get(scale).image[idx].compute()
                    _sdata[image_name].get(scale).image[idx] = (
                        percentile_normalization(im, lower_percentile, upper_percentile) * np.iinfo(np.uint16).max
                    ).astype(np.uint16)

        # get relevant segmentation masks
        if mask_names is None:
            masks = []
            if self.filehandler.nuc_seg_status:
                masks.append("seg_all_nucleus")
            if self.filehandler.cyto_seg_status:
                masks.append("seg_all_cytosol")

            if len(masks) == 0:
                raise ValueError("No segmentation masks found in the sdata object.")
        else:
            for mask in mask_names:
                if mask not in _sdata:
                    raise ValueError(f"Mask {mask} not found in the spatialdata object.")
            masks = mask_names

        # create plot
        fig, axs = plt.subplots(1, len(masks) + 1, figsize=(8 * (len(masks) + 1), 8))
        plot_segmentation_mask(
            _sdata,
            masks,
            ax=axs[0],
            title="overlayed",
            title_fontsize=title_fontsize,
            line_width=line_width,
            show_fig=False,
        )

        for mask in masks:
            idx = masks.index(mask)
            if "nucleus" in mask:
                channel = [0]
                name = "Nucleus Mask"
            elif "cytosol" in mask:
                channel = [1]
                name = "Cytosol Mask"
            else:
                channel = list(range(len(channel_names)))
                name = mask

            plot_segmentation_mask(
                _sdata,
                [mask],
                selected_channels=channel,
                ax=axs[idx + 1],
                title=name,
                title_fontsize=title_fontsize,
                line_width=line_width,
                show_fig=False,
            )

        fig.tight_layout()

        if return_fig:
            return fig
        else:
            plt.show()
            return None

    def plot_single_cell_images(
        self,
        n_cells: int | None = None,
        cell_ids: list[int] | None = None,
        show_cell_ids: bool = True,
        select_channel: int | list[int] | None = None,
        cmap="viridis",
        return_fig: bool = False,
    ) -> None | Figure:
        """Visualize the extracted single-cell images.

        Uses the function `scportrait.plotting.h5sc.cell_grid` to plot the images providing useful default values.
        For further customization of this plot please refer to the documentation of the function.

        Args:
            n_cells: Number of cells to be plotted. If not set, will proceed with default values.
            cell_ids: List of cell ids to be plotted. If not set, will randomly select cells from the dataset.
            show_cell_ids: If set to ``True``, the cell ids will be shown on the plot.
            select_channel: List of channel names or indices to be plotted. If not set, all available channels will be plotted.
            cmap: Colormap to be used for the plot. Default is ``viridis``.
            return_fig: If set to ``True``, the function returns the figure object instead of displaying it.
        """

        # import optional required dependencies for this function
        from scportrait.plotting.h5sc import cell_grid

        if cell_ids is not None:
            assert n_cells is None, "n_cells and cell_ids cannot be set at the same time."
        if n_cells is not None:
            assert cell_ids is None, "n_cells and cell_ids cannot be set at the same time."

        return cell_grid(
            self.h5sc,
            n_cells=n_cells,
            cell_ids=cell_ids,
            show_cell_id=show_cell_ids,
            select_channel=select_channel,
            cmap=cmap,
            return_fig=return_fig,
        )

    #### Functions to load input data ####
    def load_input_from_array(
        self, array: np.ndarray, channel_names: list[str] = None, overwrite: bool | None = None, remap: list[int] = None
    ) -> None:
        """Load input image from a numpy array.

        In the array the channels should be specified in the following order: nucleus, cytosol other channels.

        Args:
            array (np.ndarray): Input image as a numpy array.
            channel_names: List of channel names. Default is ``["channel_0", "channel_1", ...]``.
            overwrite (bool, None, optional): If set to ``None``, will read the overwrite value from the associated project.
                Otherwise can be set to a boolean value to override project specific settings for image loading.
            remap: List of integers that can be used to shuffle the order of the channels. For example ``[1, 0, 2]`` to invert the first two channels. Default is ``None`` in which case no reordering is performed.
                This transform is also applied to the channel names.
        Returns:
            None: Image is written to the project associated sdata object.

            The input image can be accessed using the project object::

                    project.input_image

        Examples:
            Load input images from tif files and attach them to an scportrait project::

                from scportrait.pipeline.project import Project

                project = Project("path/to/project", config_path="path/to/config.yml", overwrite=True, debug=False)
                array = np.random.rand(3, 1000, 1000)
                channel_names = ["cytosol", "nucleus", "other_channel"]
                project.load_input_from_array(array, channel_names=channel_names, remap=[1, 0, 2])

        """
        # check if an input image was already loaded if so throw error if overwrite = False

        # setup overwrite
        original_overwrite = self.overwrite
        if overwrite is not None:
            self.overwrite = overwrite

        self._cleanup_sdata_object()

        # get channel names
        if channel_names is None:
            channel_names = [f"channel_{i}" for i in range(array.shape[0])]

        assert len(channel_names) == array.shape[0], "Number of channel names does not match number of input images."

        self.channel_names = channel_names

        # ensure the array is a dask array
        image = darray.from_array(array, chunks=self.DEFAULT_CHUNK_SIZE_3D)

        if remap is not None:
            image = image[remap]
            self.channel_names = [self.channel_names[i] for i in remap]

        # write to sdata object
        self.filehandler._write_image_sdata(
            image,
            channel_names=self.channel_names,
            scale_factors=[2, 4, 8],
            chunks=self.DEFAULT_CHUNK_SIZE_3D,
            image_name=self.DEFAULT_INPUT_IMAGE_NAME,
        )

        self.get_project_status()
        self.overwrite = original_overwrite  # reset to original value

    def load_input_from_tif_files(
        self,
        file_paths: list[str],
        channel_names: list[str] = None,
        crop: list[tuple[int, int]] | None = None,
        overwrite: bool | None = None,
        remap: list[int] = None,
        cache: str | None = None,
    ):
        """
        Load input image from a list of files. The channels need to be specified in the following order: nucleus, cytosol other channels.

        Args:
            file_paths: List containing paths to each channel tiff file, like
                ``["path1/img.tiff", "path2/img.tiff", "path3/img.tiff"]``
            channel_names: List of channel names. Default is ``["channel_0", "channel_1", ...]``.
            crop (None, List[Tuple], optional): When set, it can be used to crop the input image.
                The first element refers to the first dimension of the image and so on.
                For example use ``[(0,1000),(0,2000)]`` to crop the image to 1000 px height and 2000 px width from the top left corner.
            overwrite (bool, None, optional): If set to ``None``, will read the overwrite value from the associated project.
                Otherwise can be set to a boolean value to override project specific settings for image loading.
            remap: List of integers that can be used to shuffle the order of the channels. For example ``[1, 0, 2]`` to invert the first two channels. Default is ``None`` in which case no reordering is performed.
                This transform is also applied to the channel names.
            cache: path to a directory where the temporary files should be stored. Default is ``None`` then the current working directory will be used.

        Returns:
            None: Image is written to the project associated sdata object.

            The input image can be accessed using the project object::

                    project.input_image

        Examples:
            Load input images from tif files and attach them to an scportrait project::

                from scportrait.data._datasets import dataset_3
                from scportrait.pipeline.project import Project

                project = Project("path/to/project", config_path="path/to/config.yml", overwrite=True, debug=False)
                path = dataset_3()
                image_paths = [
                    f"{path}/Ch2.tif",
                    f"{path}/Ch1.tif",
                    f"{path}/Ch3.tif",
                ]
                channel_names = ["cytosol", "nucleus", "other_channel"]
                project.load_input_from_tif_files(image_paths, channel_names=channel_names, remap=[1, 0, 2])

        """

        if crop is None:
            crop = [(0, -1), (0, -1)]

        def extract_unique_parts(paths: list[str]):
            """helper function to get unique channel names from filepaths

            Parameters
            ----------
            paths : str
                _description_

            Returns
            -------
            List[str]

            """
            # Find the common base directory
            common_base = os.path.commonpath(paths)

            # Remove the common base from each path
            unique_paths = [os.path.relpath(path, common_base) for path in paths]

            unique_parts = []
            pattern = re.compile(r"(\d+)")  # Regex pattern to match numbers

            for file_name in unique_paths:
                match = pattern.search(file_name)
                if match:
                    unique_parts.append(match.group(1))  # Extract the matched number
                else:
                    unique_parts.append(file_name)  # If no match, return the whole name

            return unique_parts

        # setup overwrite
        original_overwrite = self.overwrite
        if overwrite is not None:
            self.overwrite = overwrite

        if self.config is None:
            raise ValueError("Dataset has no config file loaded")

        # check if an input image was already loaded if so throw error if overwrite = False
        self._cleanup_sdata_object()

        # save channel names
        if channel_names is None:
            channel_names = extract_unique_parts(file_paths)

        assert len(channel_names) == len(file_paths), "Number of channel names does not match number of input images."

        self.channel_names = channel_names

        # remap can be used to shuffle the order, for example [1, 0, 2] to invert the first two channels
        # default order that is expected: Nucleus channel, cell membrane channel, other channels
        if remap is not None:
            file_paths = [file_paths[i] for i in remap]
            self.channel_names = [self.channel_names[i] for i in remap]

        if cache is None:
            cache = os.getcwd()

        self._create_temp_dir(cache)

        for i, channel_path in enumerate(file_paths):
            im = imread(channel_path)

            # add automatic conversion for uint8 as this is another very common image format
            if im.dtype == np.uint8:
                im = (
                    im.astype(np.uint16) * np.iinfo(np.uint8).max
                )  # leave set to np.uint16 explicilty here as this conversion assumes going from uint8 to uint16 if the dtype is changed then this will throw a warning later ensuring that this line is fixed

            self._check_image_dtype(im)

            im = np.array(im, dtype=self.DEFAULT_IMAGE_DTYPE)[slice(*crop[0]), slice(*crop[1])]

            if i == 0:
                # define shape of required tempmmap array to read results to
                y, x = im.shape
                c = len(file_paths)

                # initialize temp array to save results to and then append to channels
                temp_image_path = tempmmap.create_empty_mmap(
                    shape=(c, y, x),
                    dtype=self.DEFAULT_IMAGE_DTYPE,
                    tmp_dir_abs_path=self._tmp_dir_path,
                )

                channels = tempmmap.mmap_array_from_path(temp_image_path)

            channels[i] = im

        channels = daskmmap.dask_array_from_path(temp_image_path)

        self.filehandler._write_image_sdata(
            channels,
            self.DEFAULT_INPUT_IMAGE_NAME,
            channel_names=self.channel_names,
            scale_factors=[2, 4, 8],
            chunks=self.DEFAULT_CHUNK_SIZE_3D,
        )

        self.overwrite = original_overwrite  # reset to original value

        # cleanup variables and temp dir
        self._clear_cache(vars_to_delete=[temp_image_path, im, channels])
        self._clear_temp_dir()

        # strange workaround that is required so that the sdata input image does not point to the dask array anymore but
        # to the image which was written to disk
        self.get_project_status()

    def load_input_from_omezarr(
        self,
        ome_zarr_path: str,
        overwrite: None | bool = None,
        channel_names: None | list[str] = None,
        remap: list[int] = None,
    ) -> None:
        """Load input image from an ome-zarr file.

        Args:
            ome_zarr_path: Path to the ome-zarr file.
            overwrite (bool, None, optional): If set to ``None``, will read the overwrite value from the associated project.
                Otherwise can be set to a boolean value to override project specific settings for image loading.
            remap: List of integers that can be used to shuffle the order of the channels. For example ``[1, 0, 2]`` to invert the first two channels. Default is ``None`` in which case no reordering is performed.
                This transform is also applied to the channel names.

        Returns:
            None: Image is written to the project associated sdata object.

            The input image can be accessed using the project object::

                    project.input_image

        Examples:
            Load input images from an ome-zarr file and attach them to an scportrait project::

                from scportrait.pipeline.project import Project

                project = Project("path/to/project", config_path="path/to/config.yml", overwrite=True, debug=False)
                ome_zarr_path = "path/to/ome.zarr"
                project.load_input_from_omezarr(ome_zarr_path, remap=[1, 0, 2])
        """
        # setup overwrite
        original_overwrite = self.overwrite
        if overwrite is not None:
            self.overwrite = overwrite

        # check if an input image was already loaded if so throw error if overwrite = False
        self._cleanup_sdata_object()

        # read the image data
        self.log(f"trying to read file from {ome_zarr_path}")
        image = da.from_zarr(str(ome_zarr_path), component=0)

        # Access the metadata to get channel names
        zarr_store = zarr.open(ome_zarr_path, mode="r")  # Adjust the path
        metadata = zarr_store.attrs.asdict()

        if "omero" in metadata and "channels" in metadata["omero"]:
            channels = metadata["omero"]["channels"]
            channel_names = [channel["label"] for channel in channels]
        else:
            if len(image.shape) == 3:
                _, _, n_channels = image.shape
            elif len(image.shape) == 2:
                n_channels = 1
            channel_names = [f"channel_{i}" for i in range(n_channels)]

        self.channel_names = channel_names

        if remap is not None:
            image = image[remap]
            self.channel_names = [self.channel_names[i] for i in remap]

        # write to sdata object
        self.filehandler._write_image_sdata(
            image,
            channel_names=self.channel_names,
            scale_factors=[2, 4, 8],
            chunks=self.DEFAULT_CHUNK_SIZE_3D,
            image_name=self.DEFAULT_INPUT_IMAGE_NAME,
        )

        self.get_project_status()
        self.overwrite = original_overwrite  # reset to original value

    def load_input_from_dask(self, dask_array, channel_names: list[str], overwrite: bool | None = None) -> None:
        """Load input image from a dask array.

        Args:
            dask_array: Dask array containing the input image.
            channel_names: List of channel names. Default is ``["channel_0", "channel_1", ...]``.
            overwrite (bool, None, optional): If set to ``None``, will read the overwrite value from the associated project.
                Otherwise can be set to a boolean value to override project specific settings for image loading.

        Returns:
            None: Image is written to the project associated sdata object.

            The input image can be accessed using the project object::

                    project.input_image

        Examples:
            Load input images from a dask array and attach them to an scportrait project::

                from scportrait.pipeline.project import Project

                project = Project("path/to/project", config_path="path/to/config.yml", overwrite=True, debug=False)
                dask_array = da.random.random((3, 1000, 1000))
                channel_names = ["cytosol", "nucleus", "other_channel"]
                project.load_input_from_dask(dask_array, channel_names=channel_names)

        """
        # setup overwrite
        original_overwrite = self.overwrite
        if overwrite is not None:
            self.overwrite = overwrite

        self._cleanup_sdata_object()

        assert (
            len(channel_names) == dask_array.shape[0]
        ), "Number of channel names does not match number of input images."

        self.channel_names = channel_names

        self.filehandler._write_image_sdata(
            dask_array,
            channel_names=self.channel_names,
            scale_factors=[2, 4, 8],
            chunks=self.DEFAULT_CHUNK_SIZE_3D,
            image_name=self.DEFAULT_INPUT_IMAGE_NAME,
        )

        self.get_project_status()
        self.overwrite = original_overwrite

    def load_input_from_sdata(
        self,
        sdata_path,
        input_image_name: str,
        nucleus_segmentation_name: str | None = None,
        cytosol_segmentation_name: str | None = None,
        cell_id_identifier: str | None = None,
        overwrite: bool | None = None,
        keep_all: bool = True,
        remove_duplicates: bool = True,
    ) -> None:
        """
        Load input image from a spatialdata object.

        Args:
            sdata_path: Path to the spatialdata object.
            input_image_name: Name of the element in the spatial data object containing the input image.
            nucleus_segmentation_name: Name of the element in the spatial data object containing the nucleus segmentation mask. Default is ``None``.
            cytosol_segmentation_name: Name of the element in the spatial data object containing the cytosol segmentation mask. Default is ``None``.
            cell_id_identifier: column of annotating tables that contain the values that match a segmentation mask. If not provided it will assume this column carries the same name as the segmentation mask before parsing.
            overwrite (bool, None, optional): If set to ``None``, will read the overwrite value from the associated project.
                Otherwise can be set to a boolean value to override project specific settings for image loading.
            keep_all: If set to ``True``, will keep all existing elements in the sdata object in addition to renaming the desired ones. Default is ``True``.
            remove_duplicates: If keep_all and remove_duplicates is True then only one copy of the spatialdata elements selected for use with scportrait processing steps will be kept. Otherwise, the element will be saved both under the original as well as the new name.

        Returns:
            None: Image is written to the project associated sdata object and self.sdata is updated.
        """

        # setup overwrite
        original_overwrite = self.overwrite
        if overwrite is not None:
            self.overwrite = overwrite

        # check if an input image was already loaded if so throw error if overwrite = False
        self._cleanup_sdata_object()

        # read input sdata object
        sdata_input = SpatialData.read(sdata_path)
        all_elements = [x.split("/")[1] for x in sdata_input.elements_paths_in_memory()]
        dict_elems = {self.DEFAULT_INPUT_IMAGE_NAME: sdata_input[input_image_name]}

        if nucleus_segmentation_name is not None:
            dict_elems[self.nuc_seg_name] = sdata_input[nucleus_segmentation_name]
            if remove_duplicates:
                all_elements.remove(nucleus_segmentation_name)

        if cytosol_segmentation_name is not None:
            dict_elems[self.cyto_seg_name] = sdata_input[cytosol_segmentation_name]
            if remove_duplicates:
                all_elements.remove(cytosol_segmentation_name)

        # set cell_id_identifier
        if cell_id_identifier is None:
            if nucleus_segmentation_name is not None and cytosol_segmentation_name is not None:
                cell_id_identifier = cytosol_segmentation_name
                new_name = self.cyto_seg_name
            elif nucleus_segmentation_name is not None:
                cell_id_identifier = nucleus_segmentation_name
                new_name = self.nuc_seg_name
            elif cytosol_segmentation_name is not None:
                cell_id_identifier = cytosol_segmentation_name
                new_name = self.cyto_seg_name
            else:
                cell_id_identifier = None
        else:
            if nucleus_segmentation_name is not None and cytosol_segmentation_name is not None:
                new_name = self.cyto_seg_name
            elif nucleus_segmentation_name is not None:
                new_name = self.nuc_seg_name
            elif cytosol_segmentation_name is not None:
                new_name = self.cyto_seg_name
            else:
                new_name = None

        # ensure that any annotating table objects are updated with the correct labels
        table_elements = [x.split("/")[1] for x in sdata_input.elements_paths_in_memory() if x.split("/")[1] == "table"]

        for table_elem in table_elements:
            table = sdata_input[table_elem]
            rename_columns = {}
            if self.DEFAULT_CELL_ID_NAME in table.obs:
                Warning(
                    f"Column {self.DEFAULT_CELL_ID_NAME} already exists in table. Renaming to `f{self.DEFAULT_CELL_ID_NAME}_orig` to preserve compatibility with scPortrait workflow."
                )
                rename_columns[self.DEFAULT_CELL_ID_NAME] = f"{self.DEFAULT_CELL_ID_NAME}_orig"
                self.log(
                    f"Renaming column `{self.DEFAULT_CENTERS_NAME}` to `f{self.DEFAULT_CELL_ID_NAME}_orig` in table {table_elem}"
                )
            if cell_id_identifier is not None:
                rename_columns[cell_id_identifier] = self.DEFAULT_CELL_ID_NAME
                self.log(f"Renaming column `{cell_id_identifier}` to `{DEFAULT_CELL_ID_NAME}` in table {table_elem}")
                table.uns["spatialdata_attrs"]["instance_key"] = self.DEFAULT_CELL_ID_NAME
                table.uns["spatialdata_attrs"]["region"] = new_name
                table.obs["region"] = new_name
                table.obs["region"] = table.obs["region"].astype("category")
            table.obs.rename(columns=rename_columns, inplace=True)

        if keep_all:
            shutil.rmtree(self.sdata_path, ignore_errors=True)
            for elem in all_elements:
                dict_elems[elem] = sdata_input[elem]

        sdata = SpatialData.init_from_elements(dict_elems)
        sdata.write(self.sdata_path, overwrite=True)

        # update project status
        self.get_project_status()
        _, x, y = _get_shape(sdata[self.DEFAULT_INPUT_IMAGE_NAME])

        self.overwrite = original_overwrite

        if self.nuc_seg_status:
            # check input size
            _, x_mask, y_mask = _get_shape(sdata[self.nuc_seg_name])
            assert x == x_mask and y == y_mask, "Input image and nucleus segmentation mask do not match in size."

            self.filehandler._add_centers(segmentation_label=self.nuc_seg_name)
        if self.cyto_seg_status:
            # check input size
            _, x_mask, y_mask = _get_shape(sdata[self.cyto_seg_name])
            assert x == x_mask and y == y_mask, "Input image and nucleus segmentation mask do not match in size."

            self.filehandler._add_centers(segmentation_label=self.cyto_seg_name)

        # ensure that if both an nucleus and cytosol segmentation mask are loaded that they match
        if self.nuc_seg_status and self.cyto_seg_status:
            ids_nuc = set(sdata[f"{self.DEFAULT_CENTERS_NAME}_{self.nuc_seg_name}"].index.values)
            ids_cyto = set(sdata[f"{self.DEFAULT_CENTERS_NAME}_{self.cyto_seg_name}"].index.values)
            assert ids_nuc == ids_cyto, "Nucleus and cytosol segmentation masks do not match."

        self.get_project_status()

    #### Functions to perform processing ####

    def segment(self, overwrite: bool | None = None):
        # check to ensure a method has been assigned
        if self.segmentation_f is None:
            raise ValueError("No segmentation method defined")

        self.get_project_status()
        self._check_for_interactive_session()

        # ensure that an input image has been loaded
        if not self.input_image_status:
            raise ValueError("No input image loaded. Please load an input image first.")

        # setup overwrite if specified in call
        original_overwrite = self.segmentation_f.overwrite
        if overwrite is not None:
            self.segmentation_f.overwrite = overwrite

        if self.nuc_seg_status or self.cyto_seg_status:
            if not self.segmentation_f.overwrite:
                raise ValueError("Segmentation already exists. Set overwrite=True to overwrite.")

        if self.input_image is not None:
            self.segmentation_f.process(self.input_image)

        self.get_project_status()
        self.segmentation_f.overwrite = original_overwrite  # reset to original value

    def complete_segmentation(self, overwrite: bool | None = None, force_run: bool = False):
        """If a sharded Segmentation was run but individual tiles failed to segment properly, this method can be called to repeat the segmentation on the failed tiles only.
        Already calculated segmentation masks will not be recalculated.

        Args:
            overwrite: If set to ``None``, will read the overwrite value from the associated project.
                Otherwise can be set to a boolean value to override project specific settings for image loading.
            force_run: If set to ``True``, will force complete_segmentation to run even if a finalized segmentation mask is already found in the spatialdata object.
        """
        # check to ensure a method has been assigned
        if self.segmentation_f is None:
            raise ValueError("No segmentation method defined")

        self.get_project_status()
        self._check_for_interactive_session()

        # ensure that an input image has been loaded
        if not self.input_image_status:
            raise ValueError("No input image loaded. Please load an input image first.")

        # setup overwrite if specified in call
        original_overwrite = self.segmentation_f.overwrite
        if overwrite is not None:
            self.segmentation_f.overwrite = overwrite

        if self.nuc_seg_status or self.cyto_seg_status:
            if not self.segmentation_f.overwrite:
                raise ValueError("Segmentation already exists. Set overwrite = True to overwrite.")

        if self.input_image is not None:
            self.segmentation_f.complete_segmentation(self.input_image, force_run=force_run)

        self.get_project_status()
        self.segmentation_f.overwrite = original_overwrite  # reset to original value

    def extract(
        self,
        partial=False,
        n_cells=None,
        seed: int = 42,
        overwrite: bool | None = None,
        output_folder_name: str | None = None,
    ) -> None:
        """Extract single-cell images from the input image using the defined extraction method.

        Args:
            partial: If set to ``True``, will run the extraction on a subset of the image. Default is ``False``.
            n_cells: Number of cells to extract if partial is ``True``
            seed: Seed for the random number generator during a partial extraction. Default is ``42``.
            overwrite: If set to ``None``, will read the overwrite value from the associated project.
                Otherwise can be set to a boolean value to override project specific settings for image loading
            output_folder_name: can be set to override the default output folder location.

        Returns:
            None: Single-cell images are written to HDF5 files in the project associated extraction directory. File path can be accessed via ``project.extraction_f.output_path``.
        """
        if self.extraction_f is None:
            raise ValueError("No extraction method defined")

        # ensure that a segmentation has been stored that can be extracted
        self.get_project_status()
        self._check_for_interactive_session()

        if not (self.nuc_seg_status or self.cyto_seg_status):
            raise ValueError("No nucleus or cytosol segmentation loaded. Please load a segmentation first.")

        # setup overwrite if specified in call
        if overwrite is not None:
            self.extraction_f.overwrite_run_path = overwrite

        self.extraction_f(partial=partial, n_cells=n_cells, seed=seed, output_folder_name=output_folder_name)
        self.get_project_status()

    def featurize(
        self,
        n_cells: int = 0,
        data_type: Literal["complete", "partial", "filtered"] = "complete",
        partial_seed: None | int = None,
        overwrite: bool | None = None,
    ):
        if self.featurization_f is None:
            raise ValueError("No featurization method defined")

        self.get_project_status()
        self._check_for_interactive_session()

        # check that prerequisits are fullfilled to featurize cells
        assert self.featurization_f is not None, "No featurization method defined."
        assert (
            self.nuc_seg_status or self.cyto_seg_status
        ), "No nucleus or cytosol segmentation loaded. Please load a segmentation first."
        assert self.extraction_status, "No single cell data extracted. Please extract single cell data first."

        extraction_dir = self.extraction_f.get_directory()

        if data_type == "complete":
            cells_path = f"{extraction_dir}/data/{DEFAULT_EXTRACTION_FILE}"

        if data_type == "partial":
            partial_runs = [x for x in os.listdir(extraction_dir) if x.startswith("partial_data")]
            selected_runs = [x for x in partial_runs if f"ncells_{n_cells}" in x]

            if len(selected_runs) == 0:
                raise ValueError(f"No partial data found for n_cells = {n_cells}.")

            if len(selected_runs) > 1:
                if partial_seed is None:
                    raise ValueError(
                        f"Multiple partial data runs found for n_cells = {n_cells} with varying seed number. Please select one by specifying partial_seed."
                    )
                else:
                    selected_run = [x for x in selected_runs if f"seed_{partial_seed}" in x]
                    if len(selected_run) == 0:
                        raise ValueError(f"No partial data found for n_cells = {n_cells} and seed = {partial_seed}.")
                    else:
                        cells_path = f"{extraction_dir}/{selected_run[0]}/{self.DEFAULT_EXTRACTION_FILE}"
            else:
                cells_path = f"{extraction_dir}/{selected_runs[0]}/{self.DEFAULT_EXTRACTION_FILE}"

        if data_type == "filtered":
            raise ValueError("Filtered data not yet implemented.")

        print("Using extraction directory:", cells_path)

        # setup overwrite if specified in call
        if overwrite is not None:
            self.featurization_f.overwrite_run_path = overwrite

        # update the number of masks that are available in the segmentation object
        self.featurization_f.n_masks = sum([self.nuc_seg_status, self.cyto_seg_status])
        self.featurization_f.data_type = data_type

        self.featurization_f(cells_path, size=n_cells)

        self.get_project_status()

    def select(
        self,
        cell_sets: list[dict],
        calibration_marker: np.ndarray | None = None,
        name: str | None = None,
    ):
        """
        Select specified classes using the defined selection method.
        """

        if self.selection_f is None:
            raise ValueError("No selection method defined")

        self.get_project_status()
        self._check_for_interactive_session()

        if not self.nuc_seg_status and not self.cyto_seg_status:
            raise ValueError("No nucleus or cytosol segmentation loaded. Please load a segmentation first.")

        assert len(self.sdata._shared_keys) > 0, "sdata object is empty."

        self.selection_f(
            cell_sets=cell_sets,
            calibration_marker=calibration_marker,
            name=name,
        )
        self.get_project_status()



import os
import shutil

import numpy as np
import zarr
from ome_zarr.io import parse_url
from ome_zarr.scale import Scaler
from ome_zarr.writer import write_image
from spatialdata import SpatialData
from spatialdata.models import Image2DModel
from spatialdata.transformations.transformations import Identity
from tifffile import imwrite
from yattag import Doc, indent


def write_tif(image_path: str, image: np.array, dtype="uint16"):
    """_summary_

    Parameters
    ----------
    image_path : str
        absolute file path where the image should be written
    image : np.array
        image that should be saved to disk
    dtype : str, optional
        datatype to save the image as, by default "uint16"
    """
    # save using tifffile library to ensure compatibility with very large tif files
    imwrite(image_path, image.astype(dtype))


def write_ome_zarr(
    filepath: str,
    image: np.array,
    channels: list[str],
    slidename: str,
    channel_colors: list[str] = None,
    downscaling_size: int = 4,
    n_downscaling_layers: int = 4,
    chunk_size: tuple[int, int, int] = (1, 1024, 1024),
    overwrite: bool = False,
):
    """write out an image as an OME-Zarr file compatible with napari

    Parameters
    ----------
    filepath : str
        absolute path to the output file where the image should be written
    image : np.array
        image to write out
    channels : [str]
        list of channel names as strings, order needs to be the same as in the passed numpy image
    slidename : str
        string indicating the name of the slide
    channel_colors : [str], optional
        if desired a list of hex colors can be passed which will be used to set each channel color in napar
        the order needs to be as specified in `channels`
        default value is None, then a set of default colors will be used
    downscaling_size : int, optional
        _description_, by default 4
    n_downscaling_layers : int, optional
        _description_, by default 4
    chunk_size : int, int, int, optional
        _description_, by default (1, 1024, 1024)
    overwrite : bool, optional
        boolean value indicating if the outfile should be overwritten if it already exists, by default False

    """
    # delete file if it already exists
    if os.path.isdir(filepath):
        if overwrite:
            shutil.rmtree(filepath)
            print(f"Outfile {filepath} already existed and was deleted.")
        else:
            raise ValueError(f"Outfile {filepath} already exists. Set overwrite to True to delete it.")

    # setup channel colors
    if channel_colors is None:
        # hard coded channel colors for napari that we fall back on if none are specified
        channel_colors = [
            "#e60049",
            "#0bb4ff",
            "#50e991",
            "#e6d800",
            "#9b19f5",
            "#ffa300",
            "#dc0ab4",
            "#b3d4ff",
            "#00bfa0",
        ]

        # ensure that list is as long as number of channels (we just cycle through the colors)
        while len(channels) > len(channel_colors):
            channel_colors = channel_colors + channel_colors
    else:
        if len(channel_colors) != len(channels):
            raise ValueError("Number of channel colors does not match number of channels")

    # setup zarr group to initialize saving process
    loc = parse_url(filepath, mode="w").store
    group = zarr.group(store=loc)
    axes = "cyx"

    group.attrs["omero"] = {
        "name": slidename + ".ome.zarr",
        "channels": [
            {"label": channel, "color": channel_colors[i], "active": True} for i, channel in enumerate(channels)
        ],
    }

    scaler = Scaler(
        copy_metadata=False,
        downscale=downscaling_size,
        in_place=False,
        labeled=False,
        max_layer=n_downscaling_layers,
        method="nearest",
    )  # increase downscale so that large slides can also be opened in napari
    write_image(image, group=group, axes=axes, storage_options={"chunks": chunk_size}, scaler=scaler)


def write_xml(image_paths: list[str], channels: list[str], slidename: str, outdir: str = None):
    """Helper function to generate an XML for import of stitched .tifs into BIAS.

    Parameters
    ----------
    image_path : [str]
        List of absolute paths to the tif files that should be included in the XML.
    channels : [str]
        list of the channel names as they should be labelled in the XML (importat: keep the same order as in the TIF list)
    slidename : str
        string indicating the base name of the stitched images. The XML will be saved as slidename.XML
    outpath : str
        optional string indicating the path to save the XML file. If None, the XML will be saved in the same directory as the images.

    Returns
    -------
    None
        a file is written out.
    """

    # perform some basic sanity checking before proceeding

    # 1. check that the number of channels matches the number of images
    if len(image_paths) != len(channels):
        raise ValueError("Number of channels does not match the number of images")

    # 2. ensure that each channel name is included in each image name
    # this should also ensure that the tif oder and the channel order match
    for i, image_path in enumerate(image_paths):
        if channels[i] not in image_path:
            raise ValueError(f"Channel {channels[i]} is not included in image {image_path}")

    # generate the XML document for writing out
    doc, tag, text = Doc().tagtext()

    xml_header = '<?xml version="1.0" encoding="UTF-8"?>'
    doc.asis(xml_header)
    with tag("BIAS", version="1.0"):
        with tag("channels"):
            for i, channel in enumerate(channels):
                with tag("channel", id=str(i + 1)):
                    with tag("name"):
                        text(channel)
        with tag("images"):
            for i, image_path in enumerate(image_paths):
                # truncate image_path to only include the filename
                image_path = os.path.basename(image_path)
                with tag("image", url=str(image_path)):
                    with tag("channel"):
                        text(str(i + 1))

    result = indent(doc.getvalue(), indentation=" " * 4, newline="\r\n")

    # write to file
    if outdir is None:
        outdir = os.path.commonpath(image_paths)
    write_path = os.path.join(outdir, slidename + ".XML")

    with open(write_path, mode="w") as f:
        f.write(result)


def write_spatialdata(
    image_path: str,
    image: np.array,
    channel_names: list[str] = None,
    scale_factors: list[int] = None,
    overwrite: bool = False,
):
    # check if the file exists and delete if overwrite is set to True
    if scale_factors is None:
        scale_factors = [2, 4, 8]
    if os.path.exists(image_path):
        if overwrite:
            shutil.rmtree(image_path)
        else:
            raise ValueError(f"File {image_path} already exists. Set overwrite to True to delete it.")

    # check channel names
    if channel_names is None:
        channel_names = [f"channel_{i}" for i in range(image.shape[0])]
    else:
        if len(channel_names) != image.shape[0]:
            raise ValueError("Number of channel names does not match number of channels in image")

    # setup transforms
    transform_original = Identity()

    # convert image to an Image2DModel
    image = Image2DModel.parse(
        image,
        dims=["c", "y", "x"],
        c_coords=channel_names,
        scale_factors=scale_factors,
        transformations={"global": transform_original},
        rgb=False,
    )

    sdata = SpatialData(images={"input_image": image})  # memory bottleneck?
    sdata.write(image_path)



from functools import partial
from typing import TypeAlias

import dask.array as da
import numpy as np
import spatialdata
import xarray
from spatialdata import SpatialData
from spatialdata.transformations import get_transformation

from scportrait.tools.sdata.write._helper import _force_delete_object, rename_image_element
from scportrait.tools.sdata.write._write import image as write_image

ChunkSize2D: TypeAlias = tuple[int, int]
ChunkSize3D: TypeAlias = tuple[int, int, int]


def _rescale_image(
    img: da.Array, lower_quantiles: np.ndarray, IPRs: np.ndarray, dtype: np.dtype = np.uint16
) -> da.Array:
    """Helper function to apply rescaling to image

    Args:
        img: Image to be rescaled
        lower_quantiles: Lower quantiles for each scale
        IPRs: Inter-percentile ranges for each scale
        dtype: Data type of image

    Returns:
        Rescaled image with the same dtype as the input image.
    """
    img = (img.astype(float) - lower_quantiles[:, None, None]) / IPRs[:, None, None]
    img = img.clip(0, 1)
    img = img * np.iinfo(dtype).max
    return img.astype(dtype)


def percentile_normalize_image(
    sdata: SpatialData,
    image_name: str,
    lower_percentile: float | list[float] = 0.1,
    upper_percentile: float | list[float] = 99.9,
    rescaled_image_name: str | None = None,
    scale_factors: list[int] | None = None,
    overwrite: bool = True,
    chunks: ChunkSize3D | None = None,
) -> SpatialData:
    """Percentile Normalize an image in a spatialdata object.

    Args:
        sdata: SpatialData object containing the image to be percentile normalized.
        image_name: Name of the image to be percentile normalized.
        lower_percentile: Lower percentile for normalization. Default is 0.1.
        upper_percentile: Upper percentile for normalization. Default is 99.9.
        rescaled_image_name: Name of the rescaled image. Default is None.
        overwrite: Whether to overwrite existing data. Default is True.
        chunks: Chunk size for the image. Default is None.

    Returns:
        the SpatialData object is updated on file with a new element containing the rescaled image
        with the name "rescaled_image_name". If this is not provided, the name will be "{image_name}_rescaled".
        If the name is identical to the input image the input image will be overwritten.

    """
    if image_name not in sdata:
        raise ValueError(f"Image {image_name} not found in sdata")

    # define default values for generated rescaled image
    if rescaled_image_name is None:
        rescaled_image_name = f"{image_name}_rescaled"

    # get image and check for proper scaling
    image = sdata[image_name]

    if isinstance(image, xarray.DataTree):
        image = image.get("scale0").image
        # placeholder needs to be implemented
        # if a multiscale image is provided the scale_factors should be kept the same
        # scale_factors = .... code here ...

    elif isinstance(image, xarray.DataArray):
        image = image

    # get dtype
    image_dtype = image.data.dtype.type
    # get channel names

    # get channel names
    channel_names = image.c.values.tolist()

    # convert percentiles to quantiles
    if isinstance(lower_percentile, list) and isinstance(upper_percentile, list):
        assert len(lower_percentile) == len(upper_percentile), "Lower and upper percentiles must have the same length"
        for lower, upper in zip(lower_percentile, upper_percentile, strict=True):
            assert lower < upper, "Lower percentile must be less than upper percentile"
            assert lower >= 0 and upper <= 100, "Percentiles must be between 0 and 100"

        # convert to quantiles
        lower_quantile = np.array(lower_percentile) / 100
        upper_quantile = np.array(upper_percentile) / 100

        # calculate quantiles for a specific scale for each channel
        assert len(image.c) == len(lower_quantile), "Number of channels must match number of quantiles"

        lower_quantiles = np.zeros((len(image.c),), dtype=image_dtype)
        upper_quantiles = np.zeros((len(image.c),), dtype=image_dtype)
        IPRs = np.zeros((len(image.c),), dtype=image_dtype)

        for i, x in enumerate(zip(image.c, lower_quantile, upper_quantile, strict=True)):
            c, lower, upper = x
            channel = image.sel(c=c)
            lower_quantiles[i] = channel.quantile(lower, dim=["x", "y"]).compute().values
            upper_quantiles[i] = channel.quantile(upper, dim=["x", "y"]).compute().values
            IPRs[i] = upper_quantiles[i] - lower_quantiles[i]

    elif isinstance(lower_percentile, float | int) and isinstance(upper_percentile, float | int):
        assert lower_percentile < upper_percentile, "Lower percentile must be less than upper percentile"
        assert lower_percentile >= 0 and upper_percentile <= 100, "Percentiles must be between 0 and 100"

        lower_quantile = lower_percentile / 100
        upper_quantile = upper_percentile / 100

        # calculate quantiles for specific scale
        lower_quantiles = image.quantile(lower_quantile, dim=["x", "y"]).compute().values
        upper_quantiles = image.quantile(upper_quantile, dim=["x", "y"]).compute().values
        IPRs = upper_quantiles - lower_quantiles
    else:
        raise ValueError(
            f"Lower and upper percentiles are an unsupported dtype {type(lower_percentile)} and {type(upper_percentile)}"
        )

    # apply rescaling to image
    rescale_fn = partial(_rescale_image, lower_quantiles=lower_quantiles, IPRs=IPRs, dtype=image_dtype)

    data_rescaled = da.map_blocks(rescale_fn, image.data, dtype=image_dtype)

    # get local transform
    local_transform = get_transformation(image)

    if rescaled_image_name == image_name:
        # to  overwrite the image in place we currently need to use a workaround that
        # involves writing the object multiple times as we can first delete the old object
        # after the new one has been created (since it reads lazily from this object)

        rescaled_image_name = f"{image_name}_rescaled"
        write_image(
            sdata,
            image=data_rescaled,
            image_name=rescaled_image_name,
            channel_names=channel_names,
            transform=local_transform,
            overwrite=overwrite,
            scale_factors=scale_factors,
            chunks=chunks,
        )
        # to ensure that the object is updated and is backed by the new written file otherwise the original input image can not be deleted
        sdata = spatialdata.read_zarr(sdata.path)
        _force_delete_object(sdata, image_name)
        sdata = rename_image_element(
            sdata, image_element=rescaled_image_name, new_element_name=image_name
        )  # rename directory on disk instead of rewriting to improve performance

    else:
        # write rescaled image back to sdata object
        write_image(
            sdata,
            image=data_rescaled,
            image_name=rescaled_image_name,
            channel_names=channel_names,
            transform=local_transform,
            overwrite=overwrite,
            scale_factors=scale_factors,
            chunks=chunks,
        )
        sdata = spatialdata.read_zarr(
            sdata.path
        )  # to ensure that the object is updated and is backed by the new written file

    return sdata



import random

import matplotlib.pyplot as plt
import numpy as np
import torch
import torchvision.transforms as T
import torchvision.transforms.functional as TF


class RandomRotation:
    """Randomly rotate input image in x-degree steps (default is 90).

    Args:
        choices: number of possible rotations
        include_zero: include 0 degree rotation in the choices
    """

    def __init__(self, choices: int = 4, include_zero: bool = True):
        angles = np.linspace(0, 360, choices + 1)
        angles = angles[:-1]

        if not include_zero:
            delta = (360 - angles[-1]) / 2
            angles = angles + delta

        self.choices = angles.tolist()

    def __call__(self, tensor):
        angle = random.choice(self.choices)
        return TF.rotate(tensor, angle)


class GaussianNoise:
    """Add Gaussian noise to the input tensor.

    Args:
        sigma: Strength of the added noise. If a range is provided a random value will be sampled form this range.
        channels_to_exclude: List of channel indices to exclude from noise addition.
        mean: Mean of the Gaussian distribution from which the noise is sampled.
        std: Standard deviation of the Gaussian distribution from which the noise is sampled.
        deep_debug: If True, the input tensor, the sampled noise, and the transformed tensor are plotted for debugging.
    """

    def __init__(
        self,
        sigma: float | tuple[float, float] = 0.1,
        mean: float = 0.5,
        std: float = 1,
        channels_to_exclude: list[int] | None = None,
        deep_debug: bool = False,
    ):
        if channels_to_exclude is None:
            channels_to_exclude = []

        if isinstance(sigma, tuple):
            assert isinstance(sigma[0], float) and isinstance(sigma[1], float), "sigma must be a tuple of two floats."
            assert sigma[0] < sigma[1], "sigma min must be less than sigma max."
            self.sigma_sample = True
            self.sigma_range = sigma
        elif isinstance(sigma, float):
            assert sigma >= 0, "sigma must be a positive float."
            self.sigma_sample = False
            self.sigma = sigma

        self.mean = mean
        self.std = std
        self.channels = channels_to_exclude or []
        self.deep_debug = deep_debug  # can be set to true for debugging purposes

    def __call__(self, tensor: torch.Tensor) -> torch.Tensor:
        # randomly sample sigma from a uniform distribution
        if self.sigma_sample:
            lower, upper = self.sigma_range
            sigma = random.uniform(lower, upper)
        else:
            sigma = self.sigma

        scale = sigma * tensor
        sampled_noise = (
            torch.tensor(0, dtype=torch.float32).repeat(*tensor.size()).normal_(mean=self.mean, std=self.std) * scale
        )

        # Vectorized masking of excluded channels
        if tensor.ndimension() == 3:  # (C, H, W)
            sampled_noise[self.channels, :, :] = 0
        elif tensor.ndimension() == 4:  # (N, C, H, W)
            sampled_noise[:, self.channels, :, :] = 0

        sampled_noise = torch.clamp(sampled_noise, 0, 1)  # clip to 0,1 (don't want to reset the scale of our values)

        if self.deep_debug:
            n_channels = tensor.shape[1]

            fig, axs = plt.subplots(1, n_channels, figsize=(n_channels * 2, 2))
            if n_channels == 1:
                axs.imshow(tensor[0, 0], vmin=0, vmax=1)
                axs.axis("off")
                axs.set_title("Input tensor sampled noise")
            else:
                for i in range(n_channels):
                    axs[i].imshow(tensor[0, i], vmin=0, vmax=1)
                    axs[i].axis("off")
                    axs[i].set_title(f"Input tensor index {i} sampled noise")

            fig, axs = plt.subplots(1, n_channels, figsize=(n_channels * 2, 2))
            if n_channels == 1:
                axs.imshow(sampled_noise[0, 0], vmin=0, vmax=1)
                axs.axis("off")
                axs.set_title("sampled noise")
            else:
                for i in range(n_channels):
                    axs[i].imshow(sampled_noise[0, i], vmin=0, vmax=1)
                    axs[i].axis("off")
                    axs[i].set_title(f"index {i} sampled noise")

        tensor = tensor.add_(sampled_noise)
        tensor = torch.clamp(tensor, 0, 1)  # clip to 0,1 (also after addition of noise)

        if self.deep_debug:
            n_channels = tensor.shape[1]

            fig, axs = plt.subplots(1, n_channels, figsize=(n_channels * 2, 2))
            if n_channels == 1:
                axs.imshow(tensor[0, 0], vmin=0, vmax=1)
                axs.axis("off")
                axs.set_title("Transformed tensor sampled noise")
            else:
                for i in range(n_channels):
                    axs[i].imshow(tensor[0, i], vmin=0, vmax=1)
                    axs[i].axis("off")
                    axs[i].set_title(f"Transformed tensor index {i} sampled noise")

        return tensor


class GaussianBlur:
    """Apply a gaussian blur to the input image.

    Args:
        kernel_size: list of kernel sizes to randomly select from
        sigma: tuple of min and max sigma values to randomly select from
    """

    def __init__(self, kernel_size: list[int] | None = None, sigma: tuple[float, float] | None = None):
        if kernel_size is None:
            kernel_size = [1, 1, 1, 1, 5, 5, 7, 9]
        if sigma is None:
            sigma = (0.1, 2.0)

        self.kernel_size = kernel_size
        self.sigma = sigma

    def __call__(self, tensor):
        # randomly select a kernel size and sigma to add more variation
        # kernel size of 1 does not affect the image = 50% of the time no blur is added
        # pytorch randomly selects a value from a uniform distribution between (sigma_min, sigma_max)
        kernel_size = random.choice(self.kernel_size)
        sigma = self.sigma
        blur = T.GaussianBlur(kernel_size, sigma)
        return blur(tensor)


class ChannelSelector:
    """select the channel used for prediction.

    Args:
        channels: list of channel indices to keep
        num_channels: number of channels in the input tensor
    """

    def __init__(self, channels: list[int] | None = None, num_channels: int = 5):
        if channels is None:
            channels = list(range(num_channels))
        assert np.max(channels) < num_channels, "highest channel index exceeds channel numb"
        self.channels = channels

    def __call__(self, tensor):
        return tensor[self.channels, :, :]


class ChannelMultiplier:
    """
    duplicate the input tensor

    Args:
        n_reps: how often to duplicate the tensor
    """

    def __init__(self, n_reps: int = 3):
        self.n_reps = n_reps

    def __call__(self, tensor):
        return torch.zeros((self.n_reps, 1, 1)) + tensor



# helper functions for paralellization
from collections.abc import Callable
from concurrent.futures import ThreadPoolExecutor, as_completed

from tqdm.auto import tqdm


def execute_indexed_parallel(func: Callable, *, args: list, tqdm_args: dict = None, n_threads: int = 10) -> list:
    """parallelization of function call with indexed arguments using ThreadPoolExecutor. Returns a list of results in the order of the input arguments.

    Args:
        func (Callable): _description_
        args (list): _description_
        tqdm_args (dict, optional): _description_. Defaults to None.
        n_threads (int, optional): _description_. Defaults to 10.

    Returns:
        list: containing the results of the function calls in the same order as the input arguments
    """
    if tqdm_args is None:
        tqdm_args = {"total": len(args)}
    elif "total" not in tqdm_args:
        tqdm_args["total"] = len(args)

    results = [None for _ in range(len(args))]
    with ThreadPoolExecutor(n_threads) as executor:
        with tqdm(**tqdm_args) as pbar:
            futures = {executor.submit(func, *arg): i for i, arg in enumerate(args)}
            for future in as_completed(futures):
                index = futures[future]
                results[index] = future.result()
                pbar.update(1)

    return results


def execute_parallel(func: Callable, *, args: list, tqdm_args: dict = None, n_threads: int = 10) -> None:
    """parallelization of function call with ThreadPoolExecutor.

    Args:
        func (Callable): _description_
        args (list): _description_
        tqdm_args (dict, optional): _description_. Defaults to None.
        n_threads (int, optional): _description_. Defaults to 10.

    Returns:
        None
    """
    if tqdm_args is None:
        tqdm_args = {"total": len(args)}
    elif "total" not in tqdm_args:
        tqdm_args["total"] = len(args)

    with ThreadPoolExecutor(n_threads) as executor:
        with tqdm(**tqdm_args) as pbar:
            futures = {executor.submit(func, *arg): i for i, arg in enumerate(args)}
            for _ in as_completed(futures):
                pbar.update(1)

    return None



import io
import os
import platform
import shutil
import warnings
from collections.abc import Callable
from contextlib import redirect_stdout
from functools import partial as func_partial
from pathlib import PosixPath

import h5py
import numpy as np
import pandas as pd
import pytorch_lightning as pl
import torch
from alphabase.io import tempmmap
from anndata import AnnData
from spatialdata.models import TableModel
from torchvision import transforms

from scportrait.pipeline._base import ProcessingStep
from scportrait.tools.ml.datasets import H5ScSingleCellDataset
from scportrait.tools.ml.plmodels import MultilabelSupervisedModel


class _FeaturizationBase(ProcessingStep):
    DEFAULT_DATA_LOADER = H5ScSingleCellDataset
    DEFAULT_MODEL_CLASS = MultilabelSupervisedModel
    PRETRAINED_MODEL_NAMES = [
        "autophagy_classifier",
    ]
    MASK_NAMES = ["nucleus", "cytosol"]
    LABEL: str | None = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._check_config()

        self.num_workers = self.config["dataloader_worker_number"]
        self.batch_size = self.config["batch_size"]

        self.label = self.config["label"] if "label" in self.config.keys() else self.LABEL
        self.dataset_size = None
        self.channel_selection = None
        self.inference_device = None
        self.model_class = None
        self.model = None
        self.transforms = None
        self.expected_imagesize = None
        self.data_type = None

        # containers to track metadta of single-cell image dataset
        self.n_cells: list[int] = []

        self._setup_channel_selection()

        # setup deep debugging
        self.deep_debug = False

        if "overwrite_run_path" not in self.__dict__.keys():
            self.overwrite_run_path = self.overwrite

    def _check_config(self) -> None:
        """Check if all required parameters are present in the config file."""
        assert "dataloader_worker_number" in self.config.keys(), "No dataloader_worker_number specified in config file."
        assert "batch_size" in self.config.keys(), "No batch_size specified in config file."

    def _setup_output(self) -> None:
        """Helper function to generate the output directory for the featurization results."""

        if not os.path.isdir(self.directory):
            os.makedirs(self.directory)

        if self.data_type is None:
            self.run_path = os.path.join(self.directory, self.label)
        else:
            self.run_path = os.path.join(self.directory, f"{self.data_type}_{self.label}")

        if not os.path.isdir(self.run_path):
            os.makedirs(self.run_path)
            self.log(f"Created new directory for featurization results: {self.run_path}")
        else:
            if self.overwrite:
                self.log("Overwrite flag is set, deleting existing directory for featurization results.")
                shutil.rmtree(self.run_path)
                os.makedirs(self.run_path)
                self.log(f"Created new directory for featurization results: {self.run_path}")
            elif self.overwrite_run_path:
                self.log("Overwrite flag is set, deleting existing directory for featurization results.")
                shutil.rmtree(self.run_path)
                os.makedirs(self.run_path)
                self.log(f"Created new directory for featurization results: {self.run_path}")
            else:
                raise ValueError(
                    f"Directory for featurization results already exists at {self.run_path}. Please set the overwrite flag to True if you wish to overwrite the existing directory."
                )

    def _setup_log_transform(self) -> None:
        """Setup if log transformation should be applied to the inference results."""
        if "log_transform" in self.config.keys():
            self.log_transform = self.config["log_transform"]
        else:
            self.log_transform = False  # default value

    def _setup_channel_selection(self) -> None:
        """Setup which channels should be used for inference. Default is that all channels available are used."""
        if "channel_selection" in self.config.keys():
            channel_selection = self.config["channel_selection"]
            if isinstance(channel_selection, list):
                assert all(
                    isinstance(x, int) for x in channel_selection
                ), "channel_selection should be a list of integers"
                self.channel_selection = channel_selection

            elif isinstance(channel_selection, int):
                self.channel_selection = [channel_selection]
            else:
                raise ValueError("channel_selection should be an integer or a list of integers.")

        else:
            self.channel_selection = None  # default value

    def _detect_automatic_inference_device(self) -> str:
        """Automatically detect the best inference device available on the system."""

        if torch.cuda.is_available():
            inference_device = "cuda"
        if torch.backends.mps.is_available():
            inference_device = torch.device("mps")
        else:
            inference_device = "cpu"

        return inference_device

    def _get_single_cell_datafile_specs(self) -> None:
        """Extract relevant metadata from single-cell image file(s).
        Will ensure that metadata that must be consistent across files is consistent.
        """
        if isinstance(self.extraction_file, str | PosixPath):
            with h5py.File(self.extraction_file, "r") as f:
                metadata: h5py.Dataset = f["uns"][self.DEFAULT_NAME_SINGLE_CELL_IMAGES]
                self.n_masks = metadata["n_masks"][()]
                self.n_channels = metadata["n_channels"][()]
                self.n_image_channels = metadata["n_image_channels"][()]

                # strings are encoded as bytes in HDF5 files, decode them to strings
                self.channel_names = metadata["channel_names"].asstr()[:]
                self.channel_mapping = metadata["channel_mapping"].asstr()[:]

                # variable metadata can be saved directly to self
                self.n_cells.append(metadata["n_cells"][()])

        if isinstance(self.extraction_file, list):
            # metadata that must be consistent across files
            n_channels = []
            n_image_channels = []
            n_masks = []
            channel_names = []
            channel_mapping = []

            # metadata that can be different across files -> saved directly into self

            for file in self.extraction_file:
                with h5py.File(file, "r") as f:
                    metadata = f["uns"][self.DEFAULT_NAME_SINGLE_CELL_IMAGES]
                    n_masks.append(metadata["n_masks"][()])
                    n_channels.append(metadata["n_channels"][()])
                    n_image_channels.append(metadata["n_image_channels"][()])

                    # strings are encoded as bytes in HDF5 files, decode them to strings
                    channel_names.append(metadata["channel_names"].asstr()[:])
                    channel_mapping.append(metadata["channel_mapping"].asstr()[:])

                    # variable metadata can be saved directly to self
                    self.n_cells.append(metadata["n_cells"][()])

            # check to ensure that metadata that must be consistent between datasets is
            assert (x == n_masks[0] for x in n_masks), "number of masks are not consistent over all passed inputfiles."
            assert (
                x == n_channels[0] for x in n_channels
            ), "number of channels are not consistent over all passed input files."
            assert (
                x == n_image_channels[0] for x in n_image_channels
            ), "number of image channels are not consistent over all passed input files."
            assert (
                x == channel_mapping[0] for x in channel_mapping
            ), "channel mapping is not consistent over all passed input files."
            assert (
                x == channel_names[0] for x in channel_names
            ), "channel names are not consistent over all passed input files."

            # set variable names after assertions have passed to the first instance of each value
            self.n_masks = n_masks[0]
            self.n_channels = n_channels[0]
            self.n_image_channels = n_image_channels[0]
            self.channel_names = channel_names[0]
            self.channel_mapping = channel_mapping[0]

        # get names for masks and image channels seperately
        self.mask_names = [
            name
            for name, identifer in zip(self.channel_names, self.channel_mapping, strict=False)
            if identifer == "mask"
        ]
        self.mask_locs = [i for i, identifer in enumerate(self.channel_mapping) if identifer == "mask"]
        self.image_channel_names = [
            name
            for name, identifer in zip(self.channel_names, self.channel_mapping, strict=False)
            if identifer == "image_channel"
        ]
        self.image_channel_locs = [
            i for i, identifer in enumerate(self.channel_mapping) if identifer == "image_channel"
        ]

    def _setup_inference_device(self) -> None:
        """
        Configure the featurization run to use the specified inference device.
        If no device is specified, the device is automatically detected.
        """

        if "inference_device" in self.config.keys():
            self.inference_device = self.config["inference_device"]

            # check that the selected inference device is also available
            if self.inference_device in ("cuda", torch.device("cuda")):
                if not torch.cuda.is_available():
                    if torch.backends.mps.is_available():
                        self.log(
                            "CUDA specified in config file but CUDA not available on system, switching to MPS for inference."
                        )
                        self.inference_device = torch.device("mps")
                    else:
                        self.log(
                            "CUDA specified in config file but CUDA not available on system, switching to CPU for inference."
                        )
                        self.inference_device = "cpu"
                else:
                    self.inference_device = torch.device("cuda")  # ensure that the complete device is always specified

            elif self.inference_device in ("mps", torch.device("mps")):
                if not torch.backends.mps.is_available():
                    if torch.cuda.is_available():
                        self.log(
                            "MPS specified in config file but MPS not available on system, switching to CUDA for inference."
                        )
                        self.inference_device = "cuda"
                    else:
                        self.log(
                            "MPS specified in config file but MPIS not available on system, switching to CPU for inference."
                        )
                        self.inference_device = "cpu"
                else:
                    self.inference_device = torch.device("mps")  # ensure that the complete device is always specified

            elif self.inference_device in ("automatic", "auto"):
                self.inference_device = self._detect_automatic_inference_device()
                self.log(f"Automatically configured inference device to {self.inference_device}")

            elif self.inference_device == "cpu":
                if torch.backends.mps.is_available():
                    self.log(
                        "CPU specified in config file but MPS available on system. Consider changing the device for the next run."
                    )
                if torch.cuda.is_available():
                    self.log(
                        "CPU specified in config file but CUDA available on system. Consider changing the device for the next run."
                    )
            else:
                raise ValueError(
                    "Invalid inference device specified in config file. Please use one of ['cuda', 'mps', 'cpu', 'automatic', 'auto']."
                )

        else:
            self.inference_device = self._detect_automatic_inference_device()
            self.log(f"Automatically configured inference device to {self.inference_device}")

    def _general_setup(self, dataset_paths: str | list[str], return_results: bool = False) -> None:
        """Helper function to execute all setup functions that are common to all featurization steps.

        Args:
            dataset_paths: Path to the extraction file or a list of paths.
            return_results: If True, the results are returned instead of being written to file

        Returns:
            None
        """

        self.extraction_file = dataset_paths
        if not return_results:
            self._setup_output()
        self._get_single_cell_datafile_specs()
        self._setup_log_transform()
        self._setup_inference_device()

    def _get_model_specs(self) -> None:
        """Get the model"""
        # model location
        self.network_dir = self.config["network"]

        # get hyperparameters for loading model
        if "hparams_path" in self.config.keys():
            self.hparams_path = self.config["hparams_path"]
        else:
            self.hparams_path = None

        # model loading strategy: how to select which checkpoint to load if not a specific checkpoint is specified
        if "model_loading_strategy" in self.config.keys():
            strategy = self.config["model_loading_strategy"]
            if strategy not in ("max", "min", "latest", "path"):
                raise ValueError(
                    f"Invalid model loading strategy {strategy} specified. Please use one of ['max', 'min', 'latest', 'path']"
                )

            self.model_loading_strategy = self.config["model_loading_strategy"]
        else:
            self.model_loading_strategy = "max"  # default behvaiour is that the checkpoint with the highest epoch is used, in general it is highly recommended though to pass the path to a specific checkpoint file instead

        # Initiate the pytorch Lightning model class to which the checkpoing should be loaded
        if self.model_class is None:
            if "model_class" in self.config.keys():
                self.define_model_class(eval(self.config["model_class"]))
            else:
                self.define_model_class(self.DEFAULT_MODEL_CLASS)  # default model class
        else:
            self.log(
                f"Model class already defined as {self.model_class} will not overwrite. If this behaviour was unintended please set the model class to none by executing 'project.featurization_f.model_class = None'"
            )

        if "model_type" in self.config.keys():
            self.model_type = self.config["model_type"]
        else:
            self.model_type = None

    def _get_gpu_memory_usage(self):
        """Print the current memory usage on the GPU."""
        if self.inference_device == "cpu":
            return None

        elif self.inference_device == torch.device("cuda"):
            try:
                memory_usage = []

                for i in range(torch.cuda.device_count()):
                    gpu_memory = torch.cuda.memory_reserved(i) / 1024**2  # Convert bytes to MiB
                    memory_usage.append(gpu_memory)
                results = {f"GPU_{i}": f"{memory_usage[i]} MiB" for i in range(len(memory_usage))}
                return results
            except (RuntimeError, ValueError) as e:
                print("Error:", e)
                return None

        elif self.inference_device == torch.device("mps"):
            try:
                used_memory = torch.mps.driver_allocated_memory() + torch.mps.driver_allocated_memory()
                used_memory = used_memory / 1024**2  # Convert bytes to MiB
                return {"MPS": f"{used_memory} MiB"}
            except (RuntimeError, ValueError) as e:
                print("Error:", e)
                return None

        else:
            raise ValueError("Invalid inference device specified.")

    ### Functions for model loading and setup

    def _assign_model(self, model) -> None:
        """Save the model to the featurization object."""
        self.model = model

        # check if the hparams specify an expected image size
        if "hparams" in model.__dict__.keys():
            if "expected_imagesize" in model.hparams.keys():
                self.expected_imagesize = model.hparams["expected_imagesize"]

    def define_model_class(self, model_class, force_load=False) -> None:
        if isinstance(model_class, str):
            model_class = eval(model_class)  # convert string to class by evaluating it

        # check that it is a valid model class
        if force_load:
            if not issubclass(model_class, pl.LightningModule):
                Warning(
                    "Forcing the loading of the model class despite the provided model class not being a subclass of pl.LightningModule. This can lead to unexpected behaviour."
                )
            else:
                Warning(
                    "Forcing the loading of the model class is on but the provided model class is a subclass of pl.LightningModule. Consider setting the force_load parameter to False as this is not a recommended default setup."
                )
        else:
            if not issubclass(model_class, pl.LightningModule):
                raise ValueError(
                    "The provided model class is not a subclass of pl.LightningModule. Please provide a valid model class. If you are sure you wish to proceed with this, please reload the model class with the force_load parameter set to True."
                )

        self.model_class = model_class
        self.log(f"Model class defined as {model_class}")

    def _load_pretrained_model(self, model_name: str) -> pl.LightningModule:
        """
        Load a pretrained model from the scPortrait library.

        Args:
            model_name : Name of the pretrained model to load.

        Returns:
            The loaded model.
        """

        if model_name == "autophagy_classifier":
            from scportrait.tools.ml.pretrained_models import autophagy_classifier

            model = autophagy_classifier(device=self.config["inference_device"])
            self.expected_imagesize = (128, 128)
        else:
            raise ValueError(
                f"Invalid model name {model_name} specified for pretrained model. The available pretrained models are {self.PRETRAINED_MODEL_NAMES}"
            )

        # set to eval mode and move to device
        model.eval()
        model.to(self.inference_device)

        return model

    def _load_model(
        self,
        ckpt_path,
        hparams_path: str | None = None,
        model_type: str | None = None,
    ) -> pl.LightningModule:
        """Load a model from a checkpoint file and transfer it to the inference device.

        Args:
            ckpt_path: Path to the checkpoint file.
            hparams_path: Path to the hparams file. If not provided, the hparams file is assumed to be in the same directory as the checkpoint file.
            model_type: Type of the model architecture to load. Default is None. For MultiLabelSupervisedModel, this can also be specified in the hparams file under the key model_type.

        Returns:
            The loaded model.
        """

        if self.model_class is None:
            raise ValueError("Model class not defined. Please define a model class before loading a model.")

        self.log(f"Loading model from checkpoint path: {ckpt_path}")

        # get path to hparams file

        if hparams_path is None:
            if self.debug:
                self.log(
                    "No hparams file provided, assuming hparams file is in the same directory as the checkpoint file. Trying to load from default location."
                )
            hparams_path = ckpt_path.replace(os.path.basename(ckpt_path), "hparams.yml")

            if not os.path.isfile(hparams_path):
                hparams_path = ckpt_path.replace(os.path.basename(ckpt_path), "hparams.yaml")

                if not os.path.isfile(hparams_path):
                    raise ValueError(
                        f"No hparams file found at {hparams_path}. Please provide a valid hparams file path."
                    )

            if self.debug:
                self.log(f"Loading hparams file from {hparams_path}")

        if model_type is None:
            model = self.model_class.load_from_checkpoint(
                checkpoint_path=ckpt_path,
                hparams_file=hparams_path,
                map_location=self.inference_device,
            )
        else:
            model = self.model_class.load_from_checkpoint(
                checkpoint_path=ckpt_path,
                hparams_file=hparams_path,
                map_location=self.inference_device,
                model_type=model_type,
            )

        model = model.eval()
        model.to(self.inference_device)

        if self.debug:
            self.log(f"model loaded, set to eval mode and transferred to device {self.inference_device}")

        return model

    def load_model(
        self,
        ckpt_path,
        hparams_path: str | None = None,
        model_type: str | None = None,
    ) -> None:
        model = self._load_model(ckpt_path, hparams_path, model_type)
        model.eval()
        self._assign_model(model)

    ### Functions regarding dataloading and transforms ####
    def configure_transforms(self, selected_transforms: list) -> None:
        self.transforms = transforms.Compose(selected_transforms)
        self.log(f"The following transforms were applied: {self.transforms}")

    def generate_dataloader(
        self,
        dataset_paths: str | list[str],
        dataset_labels: int | list[int] = 0,
        selected_transforms: transforms.Compose = transforms.Compose([]),
        size: int = 0,
        seed: int | None = 42,
        dataset_class=DEFAULT_DATA_LOADER,
    ) -> torch.utils.data.DataLoader:
        """Create a pytorch dataloader from the provided single-cell image dataset.

        Args:
            dataset_paths: paths to the single-cell image datasets.
            selected_transforms:  List of transforms to apply to the images.
            size (optional):  Number of cells to select from the dataset. Default is 0, which means all samples are selected.
            seed (optional): Seed for the random number generator if splitting the dataset and only using a subset. Default is 42.

        Returns:
            The generated dataloader.

        """
        # generate dataset
        self.log(f"Reading data from path: {dataset_paths}")

        assert isinstance(
            self.transforms, transforms.Compose
        ), f"Transforms should be a torchvision.transforms.Compose object but recieved {self.transforms.__class__} instead."
        t = self.transforms

        if self.expected_imagesize is not None:
            self.log(f"Expected image size is set to {self.expected_imagesize}. Resizing images to this size.")
            t = transforms.Compose([t, transforms.Resize(self.expected_imagesize)])

        if isinstance(dataset_paths, list):
            assert isinstance(
                dataset_labels, list
            ), "If multiple directories are provided, multiple labels must be provided."
            paths = dataset_paths
            dataset_labels = dataset_labels
        elif isinstance(dataset_paths, str):
            assert isinstance(
                dataset_labels, int
            ), "If only one directory is provided, only one label must be provided."
            paths = [dataset_paths]
            dataset_labels = [dataset_labels]

        f = io.StringIO()
        with redirect_stdout(f):
            dataset = dataset_class(
                dir_list=paths,
                dir_labels=dataset_labels,
                transform=t,
                return_id=True,
                select_channel=self.channel_selection,
            )

        if size > 0:
            if size > len(dataset):
                raise ValueError(
                    f"Selected size {size} is larger than the dataset size {len(dataset)}. Please select a smaller size."
                )
            if size < len(dataset):
                residual_size = len(dataset) - size
                if seed is not None:
                    self.log(f"Using a seeded generator with seed {seed} to split dataset")
                    gen = torch.Generator()
                    gen.manual_seed(seed)
                    dataset, _ = torch.utils.data.random_split(dataset, [size, residual_size], generator=gen)
                else:
                    self.log("Using a random generator to split dataset.")
                    dataset, _ = torch.utils.data.random_split(dataset, [size, residual_size])

            # randomly select n elements from the dataset to process
            dataset = torch.utils.data.Subset(dataset, range(size))

        # save length of dataset for reaccess during inference
        self.dataset_size = len(dataset)
        self.log(f"Processing dataset with {self.dataset_size} cells")

        # check operating system
        if platform.system() == "Windows":
            context = "spawn"
            num_workers = 0  # need to disable multiprocessing otherwise it will throw an error
        elif platform.system() == "Darwin":
            context = "fork"
            num_workers = self.num_workers
        elif platform.system() == "Linux":
            context = "fork"
            num_workers = self.num_workers

        # generate dataloader

        if num_workers > 0:
            dataloader = torch.utils.data.DataLoader(
                dataset,
                batch_size=self.batch_size,
                shuffle=False,
                num_workers=num_workers,
                drop_last=False,
                multiprocessing_context=context,
            )
        else:
            dataloader = torch.utils.data.DataLoader(
                dataset,
                batch_size=self.batch_size,
                shuffle=False,
                num_workers=num_workers,
                drop_last=False,
            )

        self.log(
            f"Dataloader generated with a batchsize of {self.batch_size} and {self.num_workers} workers. Dataloader contains {len(dataloader)} entries."
        )

        return dataloader

    #### Inference functions ####
    def inference(
        self,
        dataloader: torch.utils.data.DataLoader,
        model_fun: Callable,
        pooler_output: bool = False,
        column_names: list | None = None,
        out_of_memory: bool = True,
    ) -> pd.DataFrame:
        """performs inference on a specific provided model and dataloader.

        Args:
            dataloader: Dataloader containing the data to perform inference on.
            model_fun: Model function to use for inference.
            pooler_output: If True, the args are passed as a ** call and the pooler output is returned. Defaults to False.
            column_names: Column names for the results dataframe. Defaults to None.


        Returns:
            pd.DataFrame: Dataframe containing the results of the inference.
        """
        self.log(f"Started processing of {len(dataloader)} batches.")

        data_iter = iter(dataloader)
        with torch.no_grad():
            # create id to track which index positions have already been filled in the results container
            ix: int = 0

            # perform first pass to get size of the returned inference results
            x, label, class_id = next(data_iter)
            if pooler_output:
                result = model_fun(x.to(self.inference_device)).pooler_output.cpu().detach()
            else:
                result = model_fun(x.to(self.inference_device)).cpu().detach()

            # initialize a datastructure for saving the results
            n_entries, n_features = result.shape
            shape_features = (self.dataset_size, n_features)
            shape_labels = (self.dataset_size, 1)

            if out_of_memory:
                # use memory-mapped temp arrays to provide out-of-memory support
                features_path = tempmmap.create_empty_mmap(
                    shape_features, dtype=np.float32, tmp_dir_abs_path=self._tmp_dir_path
                )
                cell_ids_path = tempmmap.create_empty_mmap(
                    shape_labels, dtype=np.int64, tmp_dir_abs_path=self._tmp_dir_path
                )
                labels_path = tempmmap.create_empty_mmap(
                    shape_labels, dtype=np.int64, tmp_dir_abs_path=self._tmp_dir_path
                )

                features = tempmmap.mmap_array_from_path(features_path)
                cell_ids = tempmmap.mmap_array_from_path(cell_ids_path)
                labels = tempmmap.mmap_array_from_path(labels_path)

            else:
                # use numpy arrays
                features = np.zeros(shape_features, dtype=np.float32)
                cell_ids = np.zeros(shape_labels, dtype=np.int64)
                labels = np.zeros(shape_labels, dtype=np.int64)

            # save the results for each batch into the storage container at the specified indices
            features[ix : (ix + result.shape[0])] = result.numpy()
            cell_ids[ix : (ix + result.shape[0])] = class_id.unsqueeze(1)
            labels[ix : (ix + result.shape[0])] = label.unsqueeze(1)
            ix += result.shape[0]  # update id to track filled positions

            # add check to ensure this only runs if we have more than one batch in the dataset
            if len(dataloader) > 1:
                for i in range(len(dataloader) - 1):
                    if i % 10 == 0:
                        self.log(f"processing batch {i}")

                    x, label, class_id = next(data_iter)
                    if pooler_output:
                        result = model_fun(x.to(self.inference_device)).pooler_output.cpu().detach()
                    else:
                        result = model_fun(x.to(self.inference_device)).cpu().detach()

                    # save the results for each batch into the storage container at the specified indices
                    features[ix : (ix + result.shape[0])] = result.numpy()
                    cell_ids[ix : (ix + result.shape[0])] = class_id.unsqueeze(1)
                    labels[ix : (ix + result.shape[0])] = label.unsqueeze(1)
                    ix += result.shape[0]  # update id to track filled positions

        if self.log_transform:
            self.log("Applying log transformation to results.")
            sigma = 1e-9  # to avoid log(0)
            features = np.log(features + sigma)

        # save inferred activations / predictions
        if column_names is None:
            column_names = [f"result_{i}" for i in range(features.shape[1])]

        dataframe = pd.DataFrame(data=features, columns=column_names)
        dataframe["label"] = labels
        dataframe[self.DEFAULT_CELL_ID_NAME] = cell_ids.astype("int")

        self.log("finished processing.")

        return dataframe

    #### Results writing functions ####

    def _write_results_csv(self, results: pd.DataFrame, path: str | PosixPath) -> None:
        """Write results to a CSV file."""
        results.to_csv(path, index=False)
        self.log(f"Results saved to file: {path}")

    def _write_results_sdata(self, results: pd.DataFrame, label: str, mask_type: str = "seg_all") -> None:
        """Add results to the spatialdata object.

        Args:
            results: Results to add to the spatialdata object.
            label: Label for the results.
            mask_type: Type of mask used for the results. Defaults to "seg_all".
        """
        cell_ids = results[self.DEFAULT_CELL_ID_NAME].values.astype(self.DEFAULT_SEGMENTATION_DTYPE)
        results.drop(columns=[self.DEFAULT_CELL_ID_NAME, "label"], inplace=True)
        feature_matrix = results.to_numpy()
        var_names = results.columns
        obs_indices = results.index.astype(str)

        if self.project is not None:
            if self.project.nuc_seg_status:
                # save nucleus segmentation
                obs = pd.DataFrame()
                obs.index = obs_indices
                obs[self.DEFAULT_CELL_ID_NAME] = cell_ids
                obs["region"] = f"{mask_type}_{self.MASK_NAMES[0]}"
                obs["region"] = obs["region"].astype("category")

                table = AnnData(X=feature_matrix, var=pd.DataFrame(index=var_names), obs=obs)
                table = TableModel.parse(
                    table,
                    region=[f"{mask_type}_{self.MASK_NAMES[0]}"],
                    region_key="region",
                    instance_key=self.DEFAULT_CELL_ID_NAME,
                )

                self.filehandler._write_table_object_sdata(
                    table,
                    f"{label}_{self.MASK_NAMES[0]}",
                    overwrite=self.overwrite_run_path,
                )

            if self.project.cyto_seg_status:
                # save cytoplasm segmentation
                obs = pd.DataFrame()
                obs.index = obs_indices
                obs[self.DEFAULT_CELL_ID_NAME] = cell_ids
                obs["region"] = f"{mask_type}_{self.MASK_NAMES[1]}"
                obs["region"] = obs["region"].astype("category")

                table = AnnData(X=feature_matrix, var=pd.DataFrame(index=var_names), obs=obs)
                table = TableModel.parse(
                    table,
                    region=[f"{mask_type}_{self.MASK_NAMES[1]}"],
                    region_key="region",
                    instance_key=self.DEFAULT_CELL_ID_NAME,
                )

                self.filehandler._write_table_object_sdata(
                    table,
                    f"{label}_{self.MASK_NAMES[1]}",
                    overwrite=self.overwrite_run_path,
                )

    #### Cleanup Functions ####

    def _post_processing_cleanup(self) -> None:
        """reset all attribute values to the default parameters."""
        if self.debug:
            memory_usage = self._get_gpu_memory_usage()
            self.log(f"GPU memory before performing cleanup: {memory_usage}")

        if "dataloader" in self.__dict__.keys():
            del self.dataloader  # type: ignore

        if "models" in self.__dict__.keys():
            del self.models  # type: ignore

        if "model" in self.__dict__.keys():
            del self.model  # type: ignore

        if "overwrite_run_path" in self.__dict__.keys():
            del self.overwrite_run_path  # type: ignore

        if "n_masks" in self.__dict__.keys():
            del self.n_masks  # type: ignore

        if "data_type" in self.__dict__.keys():
            del self.data_type  # type: ignore

        # reset to init values to ensure that subsequent runs are not affected by previous runs
        self.log_transform = None
        self.channel_names = None
        self.column_names = None
        self.dataset_size = None
        self.model_class = None
        self.transforms = None
        self.channel_selection = None
        self.model = None

        self._clear_cache()

        if self.debug:
            memory_usage = self._get_gpu_memory_usage()
            self.log(f"GPU memory after performing cleanup: {memory_usage}")

        # this needs to be called after the memory usage has been assesed
        if "inference_device" in self.__dict__.keys():
            del self.inference_device


###############################################
###### DeepLearning based Classification ######
###############################################


class MLClusterClassifier(_FeaturizationBase):
    """
    Perform classification on scPortrait's single-cell image datasets using a pretrained machine learning model.

    Args:
        config : Configuration for the extraction passed over from the :class:`pipeline.Project`.
        directory: Directory for the extraction log and results. Will be created if not existing yet.
        debug : Flag used to output debug information and map images.
        overwrite : Flag used to overwrite existing results.
    """

    CLEAN_LOG = True
    DEFAULT_LOG_NAME = "processing_MLClusterClassifier.log"

    def __init__(self, *args, **kwargs):
        """ """
        super().__init__(*args, **kwargs)

        if self.CLEAN_LOG:
            self._clean_log_file()

        # checks for additional essential parameters in the config file
        assert (
            self.label is not None
        ), "'label' must be specified in the config file. This is the label used to save the results."

    def _get_network_dir(self) -> pl.LightningModule:
        if self.network_dir in self.PRETRAINED_MODEL_NAMES:
            pass
        else:
            if self.model_loading_strategy == "path":
                pass

            elif self.network_dir.endswith(".ckpt"):
                Warning(
                    "Provided network ends in .ckpt, assuming this is a complete model path. To avoid this warning in the future please set the config parameter 'model_loading_strategy' to 'path'."
                )
                self.log(
                    "Provided network ends in .ckpt, assuming this is a complete model path. To avoid this warning in the future please set the config parameter 'model_loading_strategy' to 'path'."
                )
            else:
                # then we assume this is a directory containing multiple checkpoints and we want to load a specific one based on the model_loading_strategy

                model_path = self.network_dir
                checkpoints = [file for file in os.listdir(model_path) if file.endswith(".ckpt")]
                checkpoints.sort()

                if len(checkpoints) < 1:
                    raise ValueError(f"No checkpoint files found at {model_path}")

                if len(checkpoints) == 1:
                    self.log("Only one checkpoint found in network directory.")

                    # update network_dir to the path of the checkpoint
                    self.network_dir = os.path.join(model_path, checkpoints[0])
                else:
                    if self.model_loading_strategy == "max":
                        max_epoch = max([int(file.split("epoch=")[1].split("-")[0]) for file in checkpoints])
                        max_epoch_file = [file for file in checkpoints if f"epoch={max_epoch}" in file][0]
                        self.network_dir = os.path.join(model_path, max_epoch_file)
                        self.log(
                            f"Using model selection strategy 'max', selecting model with the maximum epoch {max_epoch} from path {self.network_dir}"
                        )
                    elif self.model_loading_strategy == "min":
                        min_epoch = min([int(file.split("epoch=")[1].split("-")[0]) for file in checkpoints])
                        min_epoch_file = [file for file in checkpoints if f"epoch={min_epoch}" in file][0]
                        self.network_dir = os.path.join(model_path, min_epoch_file)
                        self.log(
                            f"Using model selection strategy 'min', selecting model with the minimum epoch {min_epoch} from path {self.network_dir}"
                        )
                    elif self.model_loading_strategy == "latest":
                        self.network_dir = os.path.join(model_path, checkpoints[-1])
                        self.log(
                            f"Using model selection strategy 'latest', selecting the latest model from path {self.network_dir}"
                        )
                    else:
                        raise ValueError(
                            f"Invalid model loading strategy {self.model_loading_strategy} specified. Please use one of ['max', 'min', 'latest'] if not provding a path to a model cpkt."
                        )

    def _setup_encoders(self) -> None:
        # extract which inferences to make from config file
        if "encoders" in self.config.keys():
            encoders = self.config["encoders"]
        else:
            encoders = ["forward"]  # default parameter

        self.models = []

        for encoder in encoders:
            if encoder == "forward":
                self.models.append(self.model.network.forward)
            if encoder == "encoder":
                self.models.append(self.model.network.encoder)

    def _setup_transforms(self) -> None:
        if self.transforms is not None:
            self.log(
                "Transforms already configured manually. Will not overwrite. If this behaviour was unintended please set the transforms to None by executing 'project.featurization_f.transforms = None'"
            )

        if "transforms" in self.config.keys():
            self.transforms = eval(self.config["transforms"])
        else:
            self.transforms = transforms.Compose([])  # default is no transforms

    def _setup(self, dataset_paths: str | list[str], return_results: bool) -> None:
        self._general_setup(dataset_paths=dataset_paths, return_results=return_results)
        self._get_model_specs()
        self._get_network_dir()

        if self.network_dir in self.PRETRAINED_MODEL_NAMES:
            model = self._load_pretrained_model(self.network_dir)
        else:
            model = self._load_model(
                ckpt_path=self.network_dir,
                hparams_path=self.hparams_path,
                model_type=self.model_type,
            )

        self._assign_model(model)

        self._setup_encoders()
        self._setup_transforms()
        self.create_temp_dir()

    def process(
        self,
        dataset_paths: str | list[str],
        dataset_labels: int | list[int] = 0,
        size: int = 0,
        return_results: bool = False,
    ) -> None | list[pd.DataFrame]:
        """
        Args:
            dataset_paths: Path(s) to the single-cell dataset files on which inference should be performed. If this class is used as part of a project processing workflow this argument will be provided automatically.
            dataset_labels: Int Label(s) for the dataset(s) provided in `dataset_paths`
            size: number of cells that should be selected for inference. Default is 0, which means all cells are selected.
            return_results: boolean value indicating if the classification results should be returned as a list of pandas DataFrames or directly written to disk.

        Returns:
            None unless `return_results` is True, then the results are returned as a list of pandas DataFrames. Otherwise, the results are written to directly to file.

        Important:
            If this class is used as part of a project processing workflow, the `Project` class will automatically provide the most recent extracted single-cell
            dataset. Therefore, only the second and third arguments need to be provided.

        Example:
            .. code-block:: python

                project.featurize()


        Note:

            The following parameters are required in the config file:

            .. code-block:: yaml

                MLClusterClassifier:

                    # channel number on which the classification should be performed
                    channel_selection: 4

                    # batch size for inference
                    batch_size: 900

                    # device on which the inference should be performed
                    inference_device: "cpu"

                    # number of workers for the dataloader
                    dataloader_worker_number: 10 #needs to be 0 if using cpu

                    # pretrained model to use for classification
                    network: "autophagy_classifier"

                    # label that should be applied to the results
                    label: "Autophagy_15h_classifier2_1"

                    # which output of the model should be returned
                    encoders: ["forward"]
        """
        self.log("Started MLClusterClassifier classification.")

        # perform setup
        self._setup(dataset_paths=dataset_paths, return_results=return_results)

        self.dataloader = self.generate_dataloader(
            dataset_paths,
            dataset_labels=dataset_labels,
            selected_transforms=self.transforms,
            size=size,
            dataset_class=self.DEFAULT_DATA_LOADER,
        )

        # perform inference
        all_results = []
        for model in self.models:
            self.log(f"Starting inference for model encoder {model.__name__}")
            results = self.inference(self.dataloader, model)

            if not return_results:
                output_name = f"inference_{model.__name__}"
                path = os.path.join(self.run_path, f"{output_name}.csv")

                self._write_results_csv(results, path)
                self._write_results_sdata(results, label=f"{self.__class__.__name__ }_{self.label}_{model.__name__}")
            else:
                all_results.append(results)

        if return_results:
            self._clear_cache()
            return all_results
        else:
            self.log(f"Results saved to file: {path}")
            # perform post processing cleanup
            if not self.deep_debug:
                self._post_processing_cleanup()
            return None


class EnsembleClassifier(_FeaturizationBase):
    """
    Perform classification on scPortrait's single-cell image datasets using an ensemble of pretrained machine learning models.

    Args:
        config : Configuration for the extraction passed over from the :class:`pipeline.Project`.
        directory: Directory for the extraction log and results. Will be created if not existing yet.
        debug : Flag used to output debug information and map images.
        overwrite : Flag used to overwrite existing results.
    """

    CLEAN_LOG = True
    DEFAULT_LOG_NAME = "processing_EnsembleClassifier.log"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.CLEAN_LOG:
            self._clean_log_file()

        # checks for additional essential parameters in the config file
        assert (
            self.label is not None
        ), "'label' musst be specified in the config file. This is the label used to save the results."

    def _setup_transforms(self):
        if self.transforms is not None:
            self.log(
                "Transforms already configured manually. Will not overwrite. If this behaviour was unintended please set the transforms to None by executing 'project.classification_f.transforms = None'"
            )
            return
        else:
            self.transforms = transforms.Compose([])

    def _load_models(self):
        # load models and generate ensemble
        self.model = []
        self.model_names = []

        for model_name, model_path in self.network_dir.items():
            model = self._load_model(ckpt_path=model_path, hparams_path=self.hparams_path)

            # check for hparams expected_imagesize
            if self.expected_imagesize is None:
                if "expected_imagesize" in model.hparams.keys():
                    self.expected_imagesize = model.hparams["expected_imagesize"]
            else:
                if "expected_imagesize" in model.hparams.keys():
                    if self.expected_imagesize != model.hparams["expected_imagesize"]:
                        raise ValueError("Expected image sizes of models in ensemble do not match.")

            self.model.append(model)
            self.model_names.append(model_name)

        self.log(f"Model Ensemble generated with a total of {len(self.model)} models.")
        memory_usage = self._get_gpu_memory_usage()
        self.log(f"GPU memory usage after loading models: {memory_usage}")

    def _setup(self, dataset_paths: str, return_results: bool):
        self._general_setup(dataset_paths=dataset_paths, return_results=return_results)
        self._get_model_specs()
        self._setup_transforms()

        # ensure that the network_dir is a dictionary
        if not isinstance(self.network_dir, dict):
            raise ValueError(
                "network_dir should be a dictionary containing the model names and paths to the model checkpoints."
            )

        self._load_models()
        self.create_temp_dir()

    def process(
        self, dataset_paths: str, dataset_labels: int | list[int] = 0, size: int = 0, return_results: bool = False
    ) -> None | dict:
        """
        Args:
            dataset_paths: Path(s) to the single-cell dataset files on which inference should be performed. If this class is used as part of a project processing workflow this argument will be provided automatically.
            dataset_labels: Int Label(s) for the dataset(s) provided in `dataset_paths`
            size: number of cells that should be selected for inference. Default is 0, which means all cells are selected.
            return_results: boolean value indicating if the classification results should be returned as a list of pandas DataFrames or directly written to disk.

        Returns:
            None unless `return_results` is True, then the results are returned as a list of pandas DataFrames. Otherwise, the results are written to directly

        Important:
            If this class is used as part of a project processing workflow, the first argument will be provided by the ``Project``
            class based on the previous single-cell extraction. Therefore, no parameters need to be provided

        Example:

            .. code-block:: python

                project.featurize()

        Note:

            The following parameters are required in the config file:

            .. code-block:: yaml

                EnsembleClassifier:
                    # channel number on which the classification should be performed
                    channel_selection: 4

                    #number of threads to use for dataloader
                    dataloader_worker_number: 24

                    #batch size to pass to GPU
                    batch_size: 900

                    #path to pytorch checkpoint that should be used for inference
                    networks:
                        model1: "path/to/model1/"
                        model2: "path/to/model2/"

                    #specify input size that the models expect, provided images will be rescaled to this size
                    input_image_px: 128

                    #label under which the results will be saved
                    classification_label: "Autophagy_15h_classifier1"

                    # on which device inference should be performed
                    # for speed should be "cuda"
                    inference_device: "cuda"
        """

        self.log("Starting Ensemble Classification")

        self._setup(dataset_paths=dataset_paths, return_results=return_results)

        self.dataloader = self.generate_dataloader(
            dataset_paths,
            dataset_labels=dataset_labels,
            selected_transforms=self.transforms,
            size=size,
            dataset_class=self.DEFAULT_DATA_LOADER,
        )

        # perform inference
        all_results = {}
        for model_name, model in zip(self.model_names, self.model, strict=False):
            self.log(f"Starting inference for model {model_name}")
            results = self.inference(self.dataloader, model)

            output_name = f"ensemble_inference_{model_name}"

            if not return_results:
                path = os.path.join(self.run_path, f"{output_name}.csv")

                self._write_results_csv(results, path)
                self._write_results_sdata(results, label=f"{self.__class__.__name__ }_{model_name}")
            else:
                all_results[model_name] = results

        if return_results:
            self._clear_cache()
            return all_results
        else:
            # perform post processing cleanup
            if not self.deep_debug:
                self._post_processing_cleanup()
            return None


class ConvNeXtFeaturizer(_FeaturizationBase):
    CLEAN_LOG = True
    LABEL = "ConvNeXtFeaturizer"
    input_image_size = 224
    model_path = "facebook/convnext-xlarge-224-22k"

    """
    Compute ConvNeXt features from scPortrait's single-cell image datasets.

    This class uses the pretrained ConvNeXt model available from the Huggingface transformers library to extract features from single-cell image datasets.
    To be able to use this class you will need to install the optional dependenices for the transformers library. You can do this with `pip install "scportrait[convnext]"`.

    Args:
        config : Configuration for the extraction passed over from the :class:`pipeline.Project`.
        directory: Directory for the extraction log and results. Will be created if not existing yet.
        debug : Flag used to output debug information and map images.
        overwrite : Flag used to overwrite existing results.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.CLEAN_LOG:
            self._clean_log_file()

        # assert that the correct transformers version is installed
        try:
            import transformers
        except ImportError:
            raise ImportError("transformers is not installed. Please install it via pip install transformers") from None

        assert len(self.channel_selection) in [1, 3], "channel_selection should be either 1 or 3 channels"

        # predefine label name for convnext featurizer
        self.label = f"{self.LABEL}_{'_'.join(f'Ch{n}' for n in self.channel_selection)}"

    def _load_model(self):
        # lazy imports
        from transformers import ConvNextModel

        # silence warnings from transformers that are not relevant here
        # we do actually just want to load some of the weights to access the convnext features

        model = ConvNextModel.from_pretrained(self.model_path, use_safetensors=True)
        model.eval()
        model.to(self.inference_device)

        self._assign_model(model)

    def _silence_warnings(self):
        import logging

        from transformers import logging as hf_logging

        # Create a custom filter class to suppress specific warnings from huggingfaces transformers
        class SpecificMessageFilter(logging.Filter):
            def __init__(self, suppressed_keywords):
                super().__init__()
                self.suppressed_keywords = suppressed_keywords

            def filter(self, record):
                return not any(keyword in record.getMessage() for keyword in self.suppressed_keywords)

        # Keywords to suppress
        suppressed_keywords = [
            "Some weights of the model checkpoint at facebook",
            "Could not find image processor class in the image processor config",
        ]

        transformers_logger = hf_logging.get_logger()
        for handler in transformers_logger.handlers:
            handler.addFilter(SpecificMessageFilter(suppressed_keywords))

    def _setup_transforms(self) -> None:
        # ConvNeXt expects 0-1 rescaled images of the shape (3, 224, 224)
        # since images are already rescaled we need to multiply the channel if running on
        # a single channel and resize the images to the desired range, in addition the
        # huggingface convnext model expects images to be passed in a dictionary with the key "pixel_values"
        from scportrait.tools.ml.transforms import ChannelMultiplier

        def construct_pixel_values(input_tensors):
            return {"pixel_values": input_tensors}

        if len(self.channel_selection) == 1:
            self.transforms = transforms.Compose(
                [
                    ChannelMultiplier(3),
                    transforms.Resize((self.input_image_size, self.input_image_size)),
                    # construct_pixel_values
                ]
            )
        elif len(self.channel_selection) == 3:
            self.transforms = transforms.Compose(
                [
                    transforms.Resize((self.input_image_size, self.input_image_size)),
                    # construct_pixel_values
                ]
            )
        else:
            raise ValueError("channel_selection should be either 1 or 3 channels")

    def _generate_column_names(self) -> list:
        N_CONVNEXT_FEATURES = 2048
        column_names = [f"convnext_feature_{i}" for i in range(N_CONVNEXT_FEATURES)]
        return column_names

    def _setup(self, dataset_paths: str | list[str], return_results: bool) -> None:
        self._silence_warnings()
        self._general_setup(dataset_paths=dataset_paths, return_results=return_results)
        self._load_model()
        self._setup_transforms()
        self._load_model()
        self.create_temp_dir()

    def process(
        self,
        dataset_paths: str | list[str],
        dataset_labels: int | list[int] = 0,
        size: int = 0,
        return_results: bool = False,
    ) -> None | pd.DataFrame:
        """
        Args
            dataset_paths: Path(s) to the single-cell dataset files on which inference should be performed. If this class is used as part of a project processing workflow this argument will be provided automatically.
            dataset_labels: Int Label(s) for the dataset(s) provided in `dataset_paths`
            size: number of cells that should be selected for inference. Default is 0, which means all cells are selected.
            return_results: boolean value indicating if the classification results should be returned as a list of pandas DataFrames or directly written to disk.

        Returns:
            None if return_results is False, otherwise a pandas DataFrame containing the results.

        Important:
            If this class is used as part of a project processing workflow, the first argument will be provided by the ``Project``
            class based on the previous single-cell extraction. Therefore, only the second and third arguments need to be provided.
            The Project class will automatically provide the most recent extracted single-cell dataset together with the supplied parameters.

        Example:
            .. code-block:: python

                project.featurize()

        Note:
            The following parameters are required in the config file:

            .. code-block:: yaml

                ConvNeXtFeaturizer:
                    # number of cells in a minibatch
                    batch_size: 900

                    # number of threads to use for dataloader
                    dataloader_worker_number: 10 #needs to be 0 if using cpu

                    # what device should be used for inference
                    inference_device: "auto"

                    # how the results should be saved
                    label: "ConvNeXtFeaturizer"

                    # which channels to run inference on
                    channel_selection: 4

        """

        self._setup(dataset_paths=dataset_paths, return_results=return_results)

        self.dataloader = self.generate_dataloader(
            dataset_paths,
            dataset_labels=dataset_labels,
            selected_transforms=self.transforms,
            size=size,
            dataset_class=self.DEFAULT_DATA_LOADER,
        )

        results = self.inference(
            self.dataloader, self.model, pooler_output=True, column_names=self._generate_column_names()
        )

        if return_results:
            self._clear_cache()
            return results
        else:
            output_name = "calculated_image_features"
            path = os.path.join(self.run_path, f"{output_name}.csv")

            self._write_results_csv(results, path)
            self._write_results_sdata(results, label=self.label)

            # perform post processing cleanup
            if not self.deep_debug:
                self._post_processing_cleanup()
            return None


####### CellFeaturization based on Classic Featurecalculation #######
class _cellFeaturizerBase(_FeaturizationBase):
    CLEAN_LOG = True

    # define the output column names
    MASK_NAMES = ["nucleus", "cytosol", "cytosol_only"]
    MASK_STATISTICS = ["area"]
    CHANNEL_STATISTICS = [
        "mean",
        "median",
        "quant75",
        "quant25",
        "summed_intensity",
        "summed_intensity_area_normalized",
    ]
    LABEL = "CellFeaturizer"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.CLEAN_LOG:
            self._clean_log_file()

    def _setup_transforms(self):
        if self.transforms is not None:
            self.log(
                "Transforms already configured manually. Will not overwrite. If this behaviour was unintended please set the transforms to None by executing 'project.classification_f.transforms = None'"
            )
            return

        self.transforms = transforms.Compose([])
        return

    def _generate_column_names(self, n_masks: int, channel_names: list[str]) -> list[str]:
        if n_masks == 1:
            if self.project is not None:
                self.project.get_project_status()

                if self.project.nuc_seg_status:
                    mask_name = self.MASK_NAMES[0]
                elif self.project.cyto_seg_status:
                    mask_name = self.MASK_NAMES[1]
                else:
                    raise ValueError("no segmentation mask found in sdata object.")
            else:
                warnings.warn(
                    "No project object provided. Assuming that the provided mask is a cytosol mask", stacklevel=2
                )
                mask_name = self.MASK_NAMES[1]
            mask_names = [mask_name]

        elif n_masks == 2:
            mask_names = self.MASK_NAMES

        column_names = []
        # get the mask names with the mask attributes
        for mask in mask_names:
            for mask_stat in self.MASK_STATISTICS:
                column_names.append(f"{mask}_{mask_stat}")

        for channel_name in channel_names:
            for mask in mask_names:
                for channel_stat in self.CHANNEL_STATISTICS:
                    column_names.append(f"{channel_name}_{channel_stat}_{mask}")
        return column_names

    def calculate_statistics(self, img: torch.Tensor, n_masks: int = 2):
        """
        Calculate statistics for an image batch.

        Args:
            img : Tensor containing the image batch.
            n_masks : Number of masks in the image. Masks are always the first images in the image stack. Default is 2.

        Returns:
            The calculated image statistics
        """
        N, _, _, _ = img.shape

        mask_statistics = []
        masks = []

        for i in range(n_masks):
            mask = img[:, i] > 0
            area = mask.view(N, -1).sum(1, keepdims=True)

            masks.append(mask)
            mask_statistics.append(area)

        if n_masks == 2:
            mask = masks[1] ^ masks[0]
            area = mask.view(N, -1).sum(1, keepdims=True)

            masks.append(mask)
            mask_statistics.append(area)

        channel_statistics = []

        for channel in range(n_masks, img.shape[1]):
            img_selected = img[:, channel]

            for mask in masks:
                mask[mask == 0] = torch.nan

                # apply mask to channel to only compute the statistics over the pixels that are relevant
                _img_selected = (img_selected * mask).to(
                    torch.float32
                )  # ensure we have correct dytpe for subsequent calculations

                mean = _img_selected.view(N, -1).nanmean(1, keepdim=True)
                median = _img_selected.view(N, -1).nanquantile(q=0.5, dim=1, keepdim=True)
                quant75 = _img_selected.view(N, -1).nanquantile(q=0.75, dim=1, keepdim=True)
                quant25 = _img_selected.view(N, -1).nanquantile(q=0.25, dim=1, keepdim=True)
                summed_intensity = _img_selected.view(N, -1).sum(1, keepdim=True)
                summed_intensity_area_normalized = summed_intensity / mask_statistics[-1]

                # save results
                channel_statistics.extend(
                    [
                        mean,
                        median,
                        quant75,
                        quant25,
                        summed_intensity,
                        summed_intensity_area_normalized,
                    ]
                )

        # Generate results tensor with all values and return
        items = mask_statistics + channel_statistics
        results = torch.concat(
            items,
            1,
        )

        return results


class CellFeaturizer(_cellFeaturizerBase):
    """
    Class for extracting general image features from scPortrait's single-cell image datasets.
    The extracted features are saved to a CSV file. The features are calculated on the basis of all channels.

    The features which are calculated are:

    - Area of the masks in pixels
    - Mean intensity in the regions labelled by each of the masks
    - Median intensity in the regions labelled by each of the masks
    - 75% quantile in the regions labelled by each of the masks
    - 25% quantile in the regions labelled by each of the masks
    - Summed intensity in the regions labelled by each of the masks
    - Summed intensity in the region labelled by each of the masks normalized for area

    Args:
        config : Configuration for the extraction passed over from the :class:`pipeline.Project`.
        directory: Directory for the extraction log and results. Will be created if not existing yet.
        debug : Flag used to output debug information and map images.
        overwrite : Flag used to overwrite existing results.
    """

    DEFAULT_LOG_NAME = "processing_CellFeaturizer.log"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.channel_selection = None  # ensure that all images are passed to the function

    def _setup(self, dataset_paths: str | list[str], return_results: bool):
        self._general_setup(dataset_paths=dataset_paths, return_results=return_results)
        self._setup_transforms()
        self._get_single_cell_datafile_specs()
        self.create_temp_dir()

    def process(
        self,
        dataset_paths: str | list[str],
        dataset_labels: int | list[int] = 0,
        size: int = 0,
        return_results: bool = False,
    ) -> None | pd.DataFrame:
        """
        Args:
            dataset_paths : Paths to the single-cell dataset files on which inference should be performed. If this class is used as part of a project processing workflow this argument will be provided automatically.
            dataset_labels: labels for the provided single-cell image datasets
            size : How many cells should be selected for inference. Default is 0, meaning all cells are selected.
            return_results : If True, the results are returned as a pandas DataFrame. Otherwise the results are written out to file.

        Returns:
            None if return_results is False, otherwise a pandas DataFrame containing the results.

        Important:
            If this class is used as part of a project processing workflow, the first argument will be provided by the ``Project`` class based on the previous single-cell extraction. Therefore, only the second and third argument need to be provided. The Project class will automatically provide the most recent extraction results together with the supplied parameters.

        Note:

            The following parameters are required in the config file:

            .. code-block:: yaml

                CellFeaturizer:
                    # Number of threads to use for dataloader
                    dataloader_worker_number: 0 # needs to be 0 if using CPU

                    # Batch size to pass to GPU
                    batch_size: 900

                    # On which device inference should be performed
                    # For speed should be "cuda"
                    inference_device: "cpu"

                    # Label under which the results should be saved
                    label: "all_channels"
        """
        self.log("Started CellFeaturization of all available channels.")

        # perform setup
        self._setup(dataset_paths=dataset_paths, return_results=return_results)

        self.dataloader = self.generate_dataloader(
            dataset_paths,
            dataset_labels=dataset_labels,
            selected_transforms=self.transforms,
            size=size,
            dataset_class=self.DEFAULT_DATA_LOADER,
        )

        # generate column names
        column_names = self._generate_column_names(n_masks=self.n_masks, channel_names=self.image_channel_names)

        # define inference function
        f = func_partial(self.calculate_statistics, n_masks=self.n_masks)

        results = self.inference(
            self.dataloader,
            f,
            column_names=column_names,
        )

        if return_results:
            self._clear_cache()
            return results
        else:
            output_name = "calculated_image_features"
            path = os.path.join(self.run_path, f"{output_name}.csv")

            self._write_results_csv(results, path)
            self._write_results_sdata(results, label=self.label)

            # perform post processing cleanup
            if not self.deep_debug:
                self._post_processing_cleanup()
            self._clear_cache()
            return None


class CellFeaturizer_single_channel(_cellFeaturizerBase):
    """
    Class for extracting general image features from scPortrait's single-cell image datasets.
    The extracted features are saved to a CSV file. The features are calculated on the basis of a single specified channel.

    The features which are calculated are:

    - Area of the masks in pixels
    - Mean intensity of the chosen channel in the regions labelled by each of the masks
    - Median intensity of the chosen channel in the regions labelled by each of the masks
    - 75% quantile of the chosen channel in the regions labelled by each of the masks
    - 25% quantile of the chosen channel in the regions labelled by each of the masks
    - Summed intensity of the chosen channel in the regions labelled by each of the masks
    - Summed intensity of the chosen channel in the region labelled by each of the masks normalized for area

    Args:
        config : Configuration for the extraction passed over from the :class:`pipeline.Project`.
        directory: Directory for the extraction log and results. Will be created if not existing yet.
        debug : Flag used to output debug information and map images.
        overwrite : Flag used to overwrite existing results.
    """

    DEFAULT_LOG_NAME = "processing_CellFeaturizer.log"
    LABEL = "CellFeaturizer_single_channel"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.label = f"{self.LABEL}_{'_'.join(f'Ch{n}' for n in self.channel_selection)}"

    def _setup_channel_selection(self):
        if self.n_masks == 2:
            self.channel_selection = [0, 1, self.channel_selection]
        if self.n_masks == 1:
            self.channel_selection = [0, self.channel_selection]
        return

    def _setup(self, dataset_paths: str | list[str], return_results: bool):
        self._general_setup(dataset_paths=dataset_paths, return_results=return_results)
        self._setup_channel_selection()
        self._setup_transforms()
        self._get_single_cell_datafile_specs()
        self.create_temp_dir()

    def process(
        self, dataset_paths: str | list[str], dataset_labels: int | list[int] = 0, size=0, return_results: bool = False
    ) -> None | pd.DataFrame:
        """
        Args:
            dataset_paths : Paths to the single-cell dataset files on which inference should be performed. If this class is used as part of a project processing workflow this argument will be provided automatically.
            dataset_labels: labels for the provided single-cell image datasets
            size : How many cells should be selected for inference. Default is 0, meaning all cells are selected.
            return_results : If True, the results are returned as a pandas DataFrame. Otherwise the results are written out to file.

        Returns:
            None if return_results is False, otherwise a pandas DataFrame containing the results.

        Important:
            If this class is used as part of a project processing workflow, the first argument will be provided by the ``Project`` class based on the previous single-cell extraction. Therefore, only the second and third argument need to be provided. The Project class will automatically provide the most recent extraction results together with the supplied parameters.

        Note:

            The following parameters are required in the config file:

            .. code-block:: yaml

                CellFeaturizer:
                    # Channel number on which the featurization should be performed
                    channel_selection: 4

                    # Number of threads to use for dataloader
                    dataloader_worker_number: 0 # needs to be 0 if using CPU

                    # Batch size to pass to GPU
                    batch_size: 900

                    # On which device inference should be performed
                    # For speed should be "cuda"
                    inference_device: "cpu"

                    # Label under which the results should be saved
                    label: "Ch3_Featurization"
        """
        self.log(f"Started CellFeaturization of selected channel {self.channel_selection}.")

        # perform setup
        self._setup(dataset_paths=dataset_paths, return_results=return_results)

        self.dataloader = self.generate_dataloader(
            dataset_paths,
            dataset_labels=dataset_labels,
            selected_transforms=self.transforms,
            size=size,
            dataset_class=self.DEFAULT_DATA_LOADER,
        )

        # generate column names
        channel_name = self.channel_names[self.channel_selection]
        column_names = self._generate_column_names(n_masks=self.n_masks, channel_names=[channel_name])

        # define inference function
        f = func_partial(self.calculate_statistics, n_masks=self.n_masks)

        results = self.inference(
            self.dataloader,
            f,
            column_names=column_names,
        )
        if return_results:
            self._clear_cache()
            return results
        else:
            output_name = f"calculated_image_features_Channel_{channel_name}"
            path = os.path.join(self.run_path, f"{output_name}.csv")

            self._write_results_csv(results, path)
            self._write_results_sdata(results, label=self.label)

            # perform post processing cleanup
            if not self.deep_debug:
                self._post_processing_cleanup()
            self._clear_cache()
            return None



from scportrait.processing.images._image_processing import downsample_img_padding, percentile_normalization
from scportrait.processing.images._zstack_compression import EDF, maximum_intensity_projection

__all__ = ["downsample_img_padding", "percentile_normalization", "EDF", "maximum_intensity_projection"]
